"""The main API for the ``wombat``."""
from __future__ import annotations

import logging
import datetime
from pathlib import Path

import yaml
import pandas as pd
from attrs import Attribute, field, define
from simpy.events import Event

from wombat.core import (
    Metrics,
    FromDictMixin,
    RepairManager,
    ServiceEquipment,
    WombatEnvironment,
)
from wombat.windfarm import Windfarm
from wombat.core.port import Port
from wombat.core.library import load_yaml, library_map
from wombat.core.data_classes import convert_to_list


def _library_mapper(file_path: str | Path) -> Path:
    """Attempts to extract a default library path if one of "DINWOODIE" or "IEA_26"
    are passed, other returns ``file_path``.

    Parameters
    ----------
    file_path : str | Path
        Should be a valid file path, or one of "DINWOODIE" or "IEA_26" to indicate a
        provided library is being used.

    Returns
    -------
    str | Path
        The library path.
    """
    return Path(library_map.get(file_path, file_path)).resolve()  # type: ignore


@define(frozen=True, auto_attribs=True)
class Configuration(FromDictMixin):
    """The ``Simulation`` configuration data class that provides all the necessary
    definitions.

    Parameters
    ----------
    name: str
        Name of the simulation. Used for logging files.
    library : str
        The data directory. See ``wombat.simulation.WombatEnvironment`` for more
        details.
    layout : str
        The windfarm layout file. See ``wombat.Windfarm`` for more details.
    service_equipment : str | list[str]
        The equpiment that will be used in the simulation. See
        ``wombat.core.ServiceEquipment`` for more details.
    weather : str
        The weather profile to be used. See ``wombat.simulation.WombatEnvironment``
        for more details.
    workday_start : int
        Starting hour for a typical work shift. Can be overridden by
        equipment-specific settings.
    workday_end : int
        Ending hour for a typical work shift. Can be overridden by
        equipment-specific settings.
    inflation_rate : float
        The annual inflation rate to be used for post-processing.
    fixed_costs : str
        The file name for the fixed costs assumptions.
    project_capacity : int | float
        The total capacity of the wind plant, in MW.
    port : dict | str | Path
        The port configuration file or dictionary that will be used to setup a
        tow-to-port repair strategy, default None.
    port_distance : int | float
        The simulation-wide daily travel distance for servicing equipment. This should
        be used as a base setting when multiple or all servicing equipment will be
        operating out of the same base location, but can be individually modified.
    start_year : int
        Start year of the simulation. The exact date will be determined by
        the first valid date of this year in ``weather``.
    end_year : int
        Final year of the simulation. The exact date will be determined by
        the last valid date of this year in ``weather``.
    SAM_settings : str
        The SAM settings file to be used for financial modeling, optional, by
        default None.
    non_operational_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the environment level, an
        undefined or later starting date will be overridden for all servicing equipment
        and any modeled port, by default None.
    non_operational_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the environment level, an
        undefined or earlier ending date will be overridden for all servicing equipment
        and any modeled port, by default None.
    reduced_speed_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the environment level, an
        undefined or later starting date will be overridden for all servicing equipment
        and any modeled port, by default None.
    reduced_speed_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the environment level, an
        undefined or earlier ending date will be overridden for all servicing equipment
        and any modeled port, by default None.
    reduced_speed : float
        The maximum operating speed during the annualized reduced speed operations.
        When defined at the environment level, an undefined or faster value will be
        overridden for all servicing equipment and any modeled port, by default 0.0.
    """

    name: str
    library: Path = field(converter=_library_mapper)
    layout: str
    service_equipment: str | list[str] = field(converter=convert_to_list)
    weather: str | pd.DataFrame
    workday_start: int = field(converter=int)
    workday_end: int = field(converter=int)
    inflation_rate: float = field(converter=float)
    project_capacity: int | float = field(converter=float)
    fixed_costs: dict | str | Path = field(default=None)
    port: dict | str | Path = field(default=None)
    start_year: int = field(default=None)
    end_year: int = field(default=None)
    SAM_settings: str = field(default=None)
    port_distance: int | float = field(default=None)
    non_operational_start: str | datetime.datetime | None = field(default=None)
    non_operational_end: str | datetime.datetime | None = field(default=None)
    reduced_speed_start: str | datetime.datetime | None = field(default=None)
    reduced_speed_end: str | datetime.datetime | None = field(default=None)
    reduced_speed: float = field(default=0.0)


@define(auto_attribs=True)
class Simulation(FromDictMixin):
    """The primary API to interact with the simulation methodologies.

    Parameters
    ----------
    library_path : str
        The path to the main data library.
    config : Configuration | dict | str
        One of the following:
         - A pre-loaded ``Configuration`` object
         - A dictionary ready to be converted to a ``Configuration`` object
         - The name of the configuration file to be loaded, that will be located at:
           ``library_path`` / config / ``config``
    """

    library_path: Path = field(converter=_library_mapper)
    config: Configuration = field()

    metrics: Metrics = field(init=False)
    windfarm: Windfarm = field(init=False)
    env: WombatEnvironment = field(init=False)
    repair_manager: RepairManager = field(init=False)
    service_equipment: list[ServiceEquipment] = field(init=False)
    port: Port = field(init=False)

    def __attrs_post_init__(self) -> None:
        """Post-initialization hook."""
        self._setup_simulation()

    @config.validator  # type: ignore
    def _create_configuration(
        self, attribute: Attribute, value: str | Path | dict | Configuration
    ) -> None:
        """Validates the configuration object and creates the ``Configuration`` object
        for the simulation.

        Raises
        ------
        TypeError
            Raised if the value provided is not able to create a valid ``Configuration``
            object
        ValueError
            Raised if ``name`` and ``config.name`` or ``library_path`` and
            ``config.library`` are not aligned.

        Returns
        -------
        Configuration
            The validated simulation configuration
        """
        if isinstance(value, (str, Path)):
            try:
                value = load_yaml(self.library_path / "project/config", value)
            except FileNotFoundError:
                value = load_yaml(self.library_path / "config", value)  # type: ignore
                logging.warning(
                    "DeprecationWarning: In v0.8, all project configurations must be"
                    " located in: '<library>/project/config/"
                )
        if isinstance(value, dict):
            value = Configuration.from_dict(value)
        if isinstance(value, Configuration):
            object.__setattr__(self, attribute.name, value)
        else:
            raise TypeError(
                "``config`` must be a dictionary, valid file path to a yaml-enocoded",
                "dictionary, or ``Configuration`` object!",
            )

        if self.config.library != self.library_path:
            raise ValueError(
                f"`library_path`: {self.library_path} and the library in `config`:"
                f" {self.config.library} do not match!"
            )

    @classmethod
    def from_config(cls, config: str | Path | dict | Configuration):
        """Creates the ``Simulation`` object only the configuration contents as either a
        full file path to the configuration file, a dictionary of the configuration
        contents, or pre-loaded ``Configuration`` object.

        Parameters
        ----------
        config : str | Path | dict | Configuration
            The simulation configuration, see ``Configuration`` for more details on the
            contents. The following is a description of the acceptable contents:

            - ``str`` : the full file path of the configuration yaml file.
            - ``dict`` : a dictionary with the requried configuration settings.
            - ``Configuration`` : a pre-created ``Configuration`` object.

        Raises
        ------
        TypeError
            Raised if ``config`` is not one of the three acceptable input types.

        Returns
        -------
        Simulation
            A ready-to-run ``Simulation`` object.
        """
        if isinstance(config, (str, Path)):
            config = Path(config).resolve()  # type: ignore
            assert isinstance(config, Path)  # mypy helper
            config = load_yaml(config.parent, config.name)
        if isinstance(config, dict):
            config = Configuration.from_dict(config)
        if not isinstance(config, Configuration):
            raise TypeError(
                "``config`` must be a dictionary or ``Configuration`` object!"
            )
        assert isinstance(config, Configuration)  # mypy helper
        # NOTE: mypy is not caught up with attrs yet :(
        return cls(config.library, config)  # type: ignore

    def _setup_simulation(self):
        """Initializes the simulation objects."""
        self.env = WombatEnvironment(
            self.config.library,
            self.config.weather,
            simulation_name=self.config.name,
            workday_start=self.config.workday_start,
            workday_end=self.config.workday_end,
            start_year=self.config.start_year,
            end_year=self.config.end_year,
            port_distance=self.config.port_distance,
            non_operational_start=self.config.non_operational_start,
            non_operational_end=self.config.non_operational_end,
            reduced_speed_start=self.config.reduced_speed_start,
            reduced_speed_end=self.config.reduced_speed_end,
            reduced_speed=self.config.reduced_speed,
        )
        self.repair_manager = RepairManager(self.env)
        self.windfarm = Windfarm(self.env, self.config.layout, self.repair_manager)

        # Create the servicing equipment and set the necessary environment variables
        self.service_equipment = []
        for service_equipment in self.config.service_equipment:
            equipment = ServiceEquipment(
                self.env, self.windfarm, self.repair_manager, service_equipment
            )
            equipment.finish_setup_with_environment_variables()
            self.service_equipment.append(equipment)

        # Create the port and add any tugboats to the available servicing equipment list
        if self.config.port is not None:
            self.port = Port(
                self.env, self.windfarm, self.repair_manager, self.config.port
            )
            self.service_equipment.extend(self.port.service_equipment_manager.items)

        if self.config.project_capacity * 1000 != round(self.windfarm.capacity, 6):
            raise ValueError(
                f"Input `project_capacity`: {self.config.project_capacity:,.6f} MW is"
                f" not equal to the sum of turbine capacities:"
                f" {self.windfarm.capacity / 1000:,.6f} MW"
            )

    def run(
        self,
        until: int | float | Event | None = None,
        create_metrics: bool = True,
        save_metrics_inputs: bool = True,
    ):
        """Calls ``WombatEnvironment.run()`` and gathers the results for
        post-processing. See ``wombat.simulation.WombatEnvironment.run`` or
        ``simpy.Environment.run`` for more details.

        Parameters
        ----------
        until : Optional[int | float | Event], optional
            When to stop the simulation, by default None. See documentation on
            ``simpy.Environment.run`` for more details.
        create_metrics : bool, optional
            If True, the metrics object will be created, and not, if False, by default
            True.
        save_metrics_inputs : bool, optional
            If True, the metrics inputs data will be saved to a yaml file, with file
            references to any larger data structures that can be reloaded later. If
            False, the data will not be saved, by default True.
        """
        self.env.run(until=until)
        if save_metrics_inputs:
            self.save_metrics_inputs()
        if create_metrics:
            self.initialize_metrics()

    def initialize_metrics(self) -> None:
        """Instantiates the ``metrics`` attribute after the simulation is run."""
        events = self.env.load_events_log_dataframe()
        operations = self.env.load_operations_log_dataframe()
        power_potential, power_production = self.env.power_production_potential_to_csv(
            windfarm=self.windfarm, operations=operations, return_df=True
        )
        substation_turbine_map = {
            s_id: {k: v.tolist() for k, v in dict.items()}
            for s_id, dict in self.windfarm.substation_turbine_map.items()
        }
        capacities = [
            self.windfarm.system(t).capacity for t in self.windfarm.turbine_id
        ]
        self.metrics = Metrics(
            data_dir=self.config.library,
            events=events,
            operations=operations,
            potential=power_potential,
            production=power_production,
            inflation_rate=self.config.inflation_rate,
            project_capacity=self.config.project_capacity,
            turbine_capacities=capacities,
            fixed_costs=self.config.fixed_costs,  # type: ignore
            substation_id=self.windfarm.substation_id.tolist(),
            turbine_id=self.windfarm.turbine_id.tolist(),
            substation_turbine_map=substation_turbine_map,
            service_equipment_names=[el.settings.name for el in self.service_equipment],
            SAM_settings=self.config.SAM_settings,
        )

    def save_metrics_inputs(self) -> None:
        """Saves the inputs for the `Metrics` initialization with either a direct
        copy of the data or a file reference that can be loaded later.
        """
        substation_turbine_map = {
            s_id: {k: v.tolist() for k, v in dict.items()}
            for s_id, dict in self.windfarm.substation_turbine_map.items()
        }
        data = {
            "data_dir": str(self.config.library),
            "events": str(self.env.events_log_fname.with_suffix(".csv")),
            "operations": str(self.env.operations_log_fname.with_suffix(".csv")),
            "potential": str(self.env.power_potential_fname),
            "production": str(self.env.power_production_fname),
            "inflation_rate": self.config.inflation_rate,
            "project_capacity": self.config.project_capacity,
            "turbine_capacities": [
                self.windfarm.system(t_id).capacity for t_id in self.windfarm.turbine_id
            ],
            "fixed_costs": self.config.fixed_costs,
            "substation_id": self.windfarm.substation_id.tolist(),
            "turbine_id": self.windfarm.turbine_id.tolist(),
            "substation_turbine_map": substation_turbine_map,
            "service_equipment_names": [
                el.settings.name for el in self.service_equipment
            ],
            "SAM_settings": self.config.SAM_settings,
        }

        with open(self.env.metrics_input_fname, "w") as f:
            yaml.dump(data, f, default_flow_style=False, sort_keys=False)



"""Turbine and turbine component shared utilities."""


from __future__ import annotations

import datetime
from math import fsum
from typing import TYPE_CHECKING, Any, Callable, Sequence
from pathlib import Path
from functools import partial, update_wrapper

import attr
import attrs
import numpy as np
import pandas as pd
from attrs import Factory, Attribute, field, define
from scipy.stats import weibull_min

from wombat.utilities.time import HOURS_IN_DAY, HOURS_IN_YEAR, parse_date


if TYPE_CHECKING:
    from wombat.core import ServiceEquipment


# Define the valid servicing equipment types
VALID_EQUIPMENT = (
    "CTV",  # Crew tranfer vessel or onsite vehicle
    "SCN",  # Small crane
    "LCN",  # Large crane
    "CAB",  # Cabling equipment
    "RMT",  # Remote reset or anything performed remotely
    "DRN",  # Drone
    "DSV",  # Diving support vessel
    "TOW",  # Tugboat or support vessel for moving a turbine to a repair facility
    "AHV",  # Anchor handling vessel, typically a tugboat, w/o trigger tow-to-port
)

# Define the valid unscheduled and valid strategies
UNSCHEDULED_STRATEGIES = ("requests", "downtime")
VALID_STRATEGIES = tuple(["scheduled"] + list(UNSCHEDULED_STRATEGIES))


def convert_to_list(
    value: Sequence | str | int | float,
    manipulation: Callable | None = None,
) -> list:
    """Convert an unknown element that could be a list or single, non-sequence element
    to a list of elements.

    Parameters
    ----------
    value : Sequence | str | int | float
        The unknown element to be converted to a list of element(s).
    manipulation: Callable | None
        A function to be performed upon the individual elements, by default None.

    Returns
    -------
    list
        The new list of elements.
    """
    if isinstance(value, (str, int, float)):
        value = [value]
    if manipulation is not None:
        return [manipulation(el) for el in value]
    return list(value)


convert_to_list_upper = partial(convert_to_list, manipulation=str.upper)
update_wrapper(convert_to_list_upper, convert_to_list)

convert_to_list_lower = partial(convert_to_list, manipulation=str.lower)
update_wrapper(convert_to_list_lower, convert_to_list)


def clean_string_input(value: str | None) -> str | None:
    """Convert a string to lower case and and removes leading and trailing white spaces.

    Parameters
    ----------
    value: str
        The user input string.

    Returns
    -------
    str
        value.lower().strip()
    """
    if value is None:
        return value
    return value.lower().strip()


def annual_date_range(
    start_day: int,
    end_day: int,
    start_month: int,
    end_month: int,
    start_year: int,
    end_year: int,
) -> np.ndarray:
    """Create a series of date ranges across multiple years between the starting date
    and ending date for each year from starting year to ending year. Will only work if
    the end year is the same as the starting year or later, and the ending month, day
    combinatoion is after the starting month, day combination.

    Parameters
    ----------
    start_day : int
        Starting day of the month.
    end_day : int
        Ending day of the month.
    start_month : int
        Starting month.
    end_month : int
        Ending month.
    start_year : int
        First year of the date range.
    end_year : int
        Last year of the date range.

    Raises
    ------
    ValueError: Raised if the ending month, day combination occur after the starting
        month, day combination.
    ValueError: Raised if the ending year is prior to the starting year.

    Yields
    ------
    np.ndarray
        A ``numpy.ndarray`` of ``datetime.date`` objects.
    """
    # Check the year bounds
    if end_year < start_year:
        raise ValueError(
            f"The end_year ({start_year}) is later than the start_year ({end_year})."
        )

    # Check the month, date combination bounds
    start = datetime.datetime(2022, start_month, start_day)
    end = datetime.datetime(2022, end_month, end_day)
    if end < start:
        raise ValueError(
            f"The starting month/day combination: {start}, is after the ending: {end}."
        )

    # Create a list of arrays of date ranges for each year
    start = datetime.datetime(1, start_month, start_day)
    end = datetime.datetime(1, end_month, end_day)
    date_ranges = [
        pd.date_range(start.replace(year=year), end.replace(year=year)).date
        for year in range(start_year, end_year + 1)
    ]

    # Return a 1-D array
    return np.hstack(date_ranges)


def annualized_date_range(
    start_date: datetime.datetime,
    start_year: int,
    end_date: datetime.datetime,
    end_year: int,
) -> np.ndarray:
    """Create an annualized list of dates based on the simulation years.

    Parameters
    ----------
    start_date : str
        The string start month and day in MM/DD or MM-DD format.
    start_year : int
        The starting year in YYYY format.
    end_date : str
        The string end month and day in MM/DD or MM-DD format.
    end_year : int
        The ending year in YYYY format.

    Returns
    -------
    set[datetime.datetime]
        The set of dates the servicing equipment is available for operations.
    """
    additional = 1 if end_date < start_date else 0
    date_list = [
        pd.date_range(
            start_date.replace(year=year),
            end_date.replace(year=year + additional),
            freq="D",
        ).date
        for year in range(start_year - additional, end_year + 1)
    ]
    dates = np.hstack(date_list)
    start = datetime.date(start_year, 1, 1)
    end = datetime.date(end_year, 12, 31)
    dates = dates[(dates >= start) & (dates <= end)]
    return dates


def convert_ratio_to_absolute(ratio: int | float, total: int | float) -> int | float:
    """Convert a proportional cost to an absolute cost.

    Parameters
    ----------
    ratio : int | float
        The proportional materials cost for a ``Maintenance`` or ``Failure``.
    total: int | float
        The turbine's replacement cost.

    Returns
    -------
    int | float
        The absolute cost of the materials.
    """
    if ratio <= 1:
        return ratio * total
    return ratio


def check_start_stop_dates(
    instance: Any, attribute: attr.Attribute, value: datetime.datetime | None
) -> None:
    """Ensure the date range start and stop points are not the same."""
    if attribute.name == "non_operational_end":
        start_date = instance.non_operational_start
        start_name = "non_operational_start"
    else:
        start_date = instance.reduced_speed_start
        start_name = "reduced_speed_start"

    if value is None:
        if start_date is None:
            return
        raise ValueError(
            "A starting date was provided, but no ending date was provided"
            f" for `{attribute.name}`."
        )

    if start_date is None:
        raise ValueError(
            "An ending date was provided, but no starting date was provided"
            f" for `{start_name}`."
        )

    if start_date == value:
        raise ValueError(
            f"Starting date (`{start_name}`={start_date} and ending date"
            f" (`{attribute.name}`={value}) cannot be the same date."
        )


def valid_hour(
    instance: Any, attribute: Attribute, value: int  # pylint: disable=W0613
) -> None:
    """Validate that the input is a valid time or null value (-1).

    Parameters
    ----------
    instance : Any
        A class object.
    attribute : Attribute
        The attribute being validated.
    value : int
        A whole number, hour of the day or -1 (null), where 24 indicates end of the day.

    Raises
    ------
    ValueError
        Raised if ``value`` is not between -1 and 24, inclusive.
    """
    if value == -1:
        pass
    elif value < 0 or value > 24:
        raise ValueError(f"Input {attribute.name} must be between 0 and 24, inclusive.")


def valid_reduction(
    instance, attribute: Attribute, value: int | float  # pylint: disable=W0613
) -> None:
    """Check to see if the reduction factor is between 0 and 1, inclusive.

    Parameters
    ----------
    value : int | float
        The input value for `speed_reduction_factor`.
    """
    if value < 0 or value > 1:
        raise ValueError(
            f"Input for {attribute.name}'s `speed_reduction_factor` must be between"
            " 0 and 1, inclusive."
        )


def greater_than_zero(
    instance,  # pylint: disable=W0613
    attribute: Attribute,  # pylint: disable=W0613
    value: int | float,
) -> None:
    """Check if an input is greater than 0.

    Parameters
    ----------
    instance : Any
        The class containing the attribute to be checked.
    attribute : Attribute
        The attribute's properties.
    value : int | float
        The user-input value for the ``attribute``.

    Raises
    ------
    ValueError
        Raised if ``value`` is less than or equal to zero.
    """
    if value <= 0:
        raise ValueError("Input must be greater than 0.")


def greater_than_equal_to_zero(
    instance,  # pylint: disable=W0613
    attribute: Attribute,  # pylint: disable=W0613
    value: int | float,
) -> None:
    """Check if an input is at least 0.

    Parameters
    ----------
    instance : Any
        The class containing the attribute to be checked.
    attribute : Attribute
        The attribute's properties.
    value : int | float
        The user-input value for the ``attribute``.

    Raises
    ------
    ValueError
        Raised if ``value`` is less than or equal to zero.
    """
    if value < 0:
        raise ValueError("Input must be at least 0.")


@define
class FromDictMixin:
    """A Mixin class to allow for kwargs overloading when a data class doesn't
    have a specific parameter definied. This allows passing of larger dictionaries
    to a data class without throwing an error.

    Raises
    ------
    AttributeError
        Raised if the required class inputs are not provided.
    """

    @classmethod
    def from_dict(cls, data: dict):
        """Map a data dictionary to an `attrs`-defined class.

        TODO: Add an error to ensure that either none or all the parameters are passed

        Parameters
        ----------
            data : dict
                The data dictionary to be mapped.

        Returns
        -------
            cls : Any
                The `attrs`-defined class.
        """
        # Get all parameters from the input dictionary that map to the class init
        kwargs = {  # type: ignore
            a.name: data[a.name]  # type: ignore
            for a in cls.__attrs_attrs__  # type: ignore
            if a.name in data and a.init  # type: ignore
        }

        # Map the inputs that must be provided:
        # 1) must be initialized
        # 2) no default value defined
        required_inputs = [  # type: ignore
            a.name  # type: ignore
            for a in cls.__attrs_attrs__  # type: ignore
            if a.init and isinstance(a.default, attr._make._Nothing)  # type: ignore
        ]
        undefined = sorted(set(required_inputs) - set(kwargs))
        if undefined:
            raise AttributeError(
                f"The class defintion for {cls.__name__} is missing the following"
                f" inputs: {undefined}"
            )
        return cls(**kwargs)  # type: ignore


@define(frozen=True, auto_attribs=True)
class Maintenance(FromDictMixin):
    """Data class to store maintenance data used in subassembly and cable modeling.

    Parameters
    ----------
    time : float
        Amount of time required to perform maintenance, in hours.
    materials : float
        Cost of materials required to perform maintenance, in USD.
    frequency : float
        Optimal number of days between performing maintenance, in days.
    service_equipment: list[str] | str
        Any combination of th following ``Equipment.capability`` options.

        - RMT: remote (no actual equipment BUT no special implementation)
        - DRN: drone
        - CTV: crew transfer vessel/vehicle
        - SCN: small crane (i.e., field support vessel)
        - LCN: large crane (i.e., heavy lift vessel)
        - CAB: cabling vessel/vehicle
        - DSV: diving support vessel
        - TOW: tugboat or towing equipment
        - AHV: anchor handling vessel (tugboat that doesn't trigger tow-to-port)
    system_value : Union[int, float]
        Turbine replacement value. Used if the materials cost is a proportional cost.
    description : str
        A short text description to be used for logging.
    operation_reduction : float
        Performance reduction caused by the failure, between (0, 1]. Defaults to 0.

        .. warning:: As of v0.7, availability is very sensitive to the usage of this
            parameter, and so it should be used carefully.

    level : int, optional
        Severity level of the maintenance. Defaults to 0.
    """

    time: float = field(converter=float)
    materials: float = field(converter=float)
    frequency: float = field(converter=float)
    service_equipment: list[str] = field(
        converter=convert_to_list_upper,  # type: ignore
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.in_(VALID_EQUIPMENT),
            iterable_validator=attrs.validators.instance_of(list),
        ),
    )
    system_value: int | float = field(converter=float)
    description: str = field(default="routine maintenance", converter=str)
    operation_reduction: float = field(default=0.0, converter=float)
    level: int = field(default=0, converter=int)
    request_id: str = field(init=False)
    replacement: bool = field(default=False, init=False)

    def __attrs_post_init__(self):
        """Convert frequency to hours (simulation time scale) and the equipment
        requirement to a list.
        """
        object.__setattr__(self, "frequency", self.frequency * HOURS_IN_DAY)
        # object.__setattr__(
        #     self,
        #     "service_equipment",
        #     convert_to_list(self.service_equipment, str.upper),
        # )
        object.__setattr__(
            self,
            "materials",
            convert_ratio_to_absolute(self.materials, self.system_value),
        )

    def assign_id(self, request_id: str) -> None:
        """Assign a unique identifier to the request.

        Parameters
        ----------
        request_id : str
            The ``wombat.core.RepairManager`` generated identifier.
        """
        object.__setattr__(self, "request_id", request_id)


@define(frozen=True, auto_attribs=True)
class Failure(FromDictMixin):
    """Data class to store failure data used in subassembly and cable modeling.

    Parameters
    ----------
    scale : float
        Weibull scale parameter for a failure classification.
    shape : float
        Weibull shape parameter for a failure classification.
    time : float
        Amount of time required to complete the repair, in hours.
    materials : float
        Cost of the materials required to complete the repair, in $USD.
    operation_reduction : float
        Performance reduction caused by the failure, between (0, 1].

        .. warning:: As of v0.7, availability is very sensitive to the usage of this
            parameter, and so it should be used carefully.

    level : int, optional
        Level of severity, will be generated in the ``ComponentData.create_severities``
        method.
    service_equipment: list[str] | str
        Any combination of the following ``Equipment.capability`` options:

        - RMT: remote (no actual equipment BUT no special implementation)
        - DRN: drone
        - CTV: crew transfer vessel/vehicle
        - SCN: small crane (i.e., field support vessel)
        - LCN: large crane (i.e., heavy lift vessel)
        - CAB: cabling vessel/vehicle
        - DSV: diving support vessel
        - TOW: tugboat or towing equipment
        - AHV: anchor handling vessel (tugboat that doesn't trigger tow-to-port)
    system_value : Union[int, float]
        Turbine replacement value. Used if the materials cost is a proportional cost.
    replacement : bool
        True if triggering the failure requires a subassembly replacement, False, if
        only a repair is necessary. Defaults to False
    description : str
        A short text description to be used for logging.
    """

    scale: float = field(converter=float)
    shape: float = field(converter=float)
    time: float = field(converter=float)
    materials: float = field(converter=float)
    operation_reduction: float = field(converter=float)
    level: int = field(converter=int)
    service_equipment: list[str] | str = field(
        converter=convert_to_list_upper,  # type: ignore
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.in_(VALID_EQUIPMENT),
            iterable_validator=attrs.validators.instance_of(list),
        ),
    )
    system_value: int | float = field(converter=float)
    replacement: bool = field(default=False, converter=bool)
    description: str = field(default="failure", converter=str)
    weibull: weibull_min = field(init=False, eq=False)
    request_id: str = field(init=False)

    def __attrs_post_init__(self):
        """Create the actual Weibull distribution and converts equipment requirements
        to a list.
        """
        object.__setattr__(self, "weibull", weibull_min(self.shape, scale=self.scale))
        object.__setattr__(
            self,
            "service_equipment",
            convert_to_list(self.service_equipment, str.upper),
        )
        object.__setattr__(
            self,
            "materials",
            convert_ratio_to_absolute(self.materials, self.system_value),
        )

    def hours_to_next_failure(self) -> float | None:
        """Sample the next time to failure in a Weibull distribution. If the ``scale``
        and ``shape`` parameters are set to 0, then the model will return ``None`` to
        cause the subassembly to timeout to the end of the simulation.

        Returns
        -------
        float | None
            Returns ``None`` for a non-modelled failure, or the time until the next
            simulated failure.
        """
        if self.scale == self.shape == 0:
            return None

        return self.weibull.rvs(size=1)[0] * HOURS_IN_YEAR

    def assign_id(self, request_id: str) -> None:
        """Assign a unique identifier to the request.

        Parameters
        ----------
        request_id : str
            The ``wombat.core.RepairManager`` generated identifier.
        """
        object.__setattr__(self, "request_id", request_id)


@define(frozen=True, auto_attribs=True)
class SubassemblyData(FromDictMixin):
    """Data storage and validation class for the subassemblies.

    Parameters
    ----------
    name : str
        Name of the component/subassembly.
    maintenance : list[dict[str, float | str]]
        List of the maintenance classification dictionaries. This will be converted
        to a list of ``Maintenance`` objects in the post initialization hook.
    failures : fict[int, dict[str, float | str]]
        Dictionary of failure classifications in a numerical (ordinal) categorization
        order. This will be converted to a dictionary of ``Failure`` objects in the
        post initialization hook.
    system_value : int | float
        Turbine's cost of replacement. Used in case percentages of turbine cost are used
        in place of an absolute cost.
    """

    name: str = field(converter=str)
    maintenance: list[Maintenance | dict[str, float | str]]
    failures: dict[int, Failure | dict[str, float | str]]
    system_value: int | float = field(converter=float)

    def __attrs_post_init__(self):
        """Convert the maintenance and failure data to ``Maintenance`` and ``Failure``
        objects, respectively.
        """
        for kwargs in self.maintenance:
            assert isinstance(kwargs, dict)
            kwargs.update({"system_value": self.system_value})
        object.__setattr__(
            self,
            "maintenance",
            [
                Maintenance.from_dict(kw) if isinstance(kw, dict) else kw
                for kw in self.maintenance
            ],
        )

        for kwargs in self.failures.values():  # type: ignore
            assert isinstance(kwargs, dict)
            kwargs.update({"system_value": self.system_value})
        object.__setattr__(
            self,
            "failures",
            {
                level: Failure.from_dict(kw) if isinstance(kw, dict) else kw
                for level, kw in self.failures.items()
            },
        )


@define(frozen=True, auto_attribs=True)
class RepairRequest(FromDictMixin):
    """Repair/Maintenance request data class.

    Parameters
    ----------
    system_id : str
        ``System.id``.
    system_name : str
        ``System.name``.
    subassembly_id : str
        ``Subassembly.id``.
    subassembly_name : str
        ``Subassembly.name``.
    severity_level : int
        ``Maintenance.level`` or ``Failure.level``.
    details : Failure | Maintenance
        The actual data class.
    cable : bool
        Indicator that the request is for a cable, by default False.
    upstream_turbines : list[str]
        The cable's upstream turbines, by default []. No need to use this if
        ``cable`` == False.
    upstream_cables : list[str]
        The cable's upstream cables, by default []. No need to use this if
        ``cable`` == False.
    """

    system_id: str = field(converter=str)
    system_name: str = field(converter=str)
    subassembly_id: str = field(converter=str)
    subassembly_name: str = field(converter=str)
    severity_level: int = field(converter=int)
    details: Failure | Maintenance = field(
        validator=attrs.validators.instance_of((Failure, Maintenance))
    )
    cable: bool = field(default=False, converter=bool, kw_only=True)
    upstream_turbines: list[str] = field(default=Factory(list), kw_only=True)
    upstream_cables: list[str] = field(default=Factory(list), kw_only=True)
    request_id: str = field(init=False)

    def assign_id(self, request_id: str) -> None:
        """Assign a unique identifier to the request.

        Parameters
        ----------
        request_id : str
            The ``wombat.core.RepairManager`` generated identifier.
        """
        object.__setattr__(self, "request_id", request_id)
        self.details.assign_id(request_id)


@define(frozen=True, auto_attribs=True)
class ServiceCrew(FromDictMixin):
    """An internal data class for the indivdual crew units that are on the servicing
    equipment.

    Parameters
    ----------
    n_day_rate: int
        Number of salaried workers.
    day_rate: float
        Day rate for salaried workers, in USD.
    n_hourly_rate: int
        Number of hourly/subcontractor workers.
    hourly_rate: float
        Hourly labor rate for subcontractors, in USD.
    """

    n_day_rate: int = field(converter=int)
    day_rate: float = field(converter=float)
    n_hourly_rate: int = field(converter=int)
    hourly_rate: float = field(converter=float)


class DateLimitsMixin:
    """Base servicing equpment dataclass. Only meant to reduce repeated code across the
    ``ScheduledServiceEquipmentData`` and ``UnscheduledServiceEquipmentData`` classes.
    """

    # MyPy type hints
    port_distance: float = field(converter=float)
    non_operational_start: datetime.datetime = field(converter=parse_date)
    non_operational_end: datetime.datetime = field(converter=parse_date)
    reduced_speed_start: datetime.datetime = field(converter=parse_date)
    reduced_speed_end: datetime.datetime = field(converter=parse_date)

    def _set_environment_shift(self, start: int, end: int) -> None:
        """Set the ``workday_start`` and ``workday_end`` to the environment's values.

        Parameters
        ----------
        start : int
            Starting hour of a workshift.
        end : int
            Ending hour of a workshift.
        """
        object.__setattr__(self, "workday_start", start)
        object.__setattr__(self, "workday_end", end)

    def _set_port_distance(self, distance: int | float | None) -> None:
        """Set ``port_distance`` from the environment's or port's variables.

        Parameters
        ----------
        distance : int | float
            The distance to port that must be traveled for servicing equipment.
        """
        if distance is None:
            return
        if distance <= 0:
            return
        if self.port_distance <= 0:  # type: ignore
            object.__setattr__(self, "port_distance", float(distance))

    def _compare_dates(
        self,
        new_start: datetime.datetime | None,
        new_end: datetime.datetime | None,
        which: str,
    ) -> tuple[datetime.datetime | None, datetime.datetime | None]:
        """Compare the orignal and newly input starting ane ending date for either the
        non-operational date range or the reduced speed date range.

        Parameters
        ----------
        new_start : datetime.datetime | None
            The overriding start date if it is an earlier date than the original.
        new_end : datetime.datetime | None
            The overriding end date if it is a later date than the original.
        which : str
            One of "reduced_speed" or "non_operational"

        Returns
        -------
        tuple[datetime.datetime | None, datetime.datetime | None]
            The more conservative date pair between the new values and original values.

        Raises
        ------
        ValueError
            Raised if an invalid value to ``which`` is provided.
        """
        if which in ("reduced_speed", "non_operational"):
            original_start = getattr(self, f"{which}_start")
            original_end = getattr(self, f"{which}_end")
        else:
            raise ValueError(
                "`which` must be one of 'reduced_speed' or 'non_operational'."
            )

        if original_start is not None:
            if new_start is not None:
                new_start = min(new_start, original_start)
            else:
                new_start = original_start
        else:
            object.__setattr__(self, f"{which}_start", new_start)

        if original_end is not None:
            if new_end is not None:
                new_end = max(new_end, original_end)
            else:
                new_end = original_end
        else:
            object.__setattr__(self, f"{which}_end", new_end)

        return new_start, new_end

    def set_non_operational_dates(
        self,
        start_date: datetime.datetime | None = None,
        start_year: int | None = None,
        end_date: datetime.datetime | None = None,
        end_year: int | None = None,
    ) -> None:
        """Create an annualized list of dates where servicing equipment should be
        unavailable. Takes a a ``start_year`` and ``end_year`` to determine how many
        years should be covered.

        Parameters
        ----------
        start_date : datetime.datetime | None
            The starting date for the annual date range of non-operational dates, by
            default None.
        start_year : int | None
            The first year that has a non-operational date range, by default None.
        end_date : datetime.datetime | None
            The ending date for the annual range of non-operational dates, by default
            None.
        end_year : int | None
            The last year that should have a non-operational date range, by default
            None.

        Raises
        ------
        ValueError
            Raised if the starting and ending dates are the same date.
        """
        start_date, end_date = self._compare_dates(
            start_date, end_date, "non_operational"
        )
        object.__setattr__(self, "non_operational_start", start_date)
        object.__setattr__(self, "non_operational_end", end_date)

        # Check that the base dates are valid
        if self.non_operational_start is None or self.non_operational_end is None:
            object.__setattr__(
                self, "non_operational_dates", np.empty(0, dtype="object")
            )
            object.__setattr__(self, "non_operational_dates_set", set())
            return

        # Check that the input year range is valid
        if not isinstance(start_year, int):
            raise ValueError(
                f"Input to `start_year`: {start_year}, must be an integer."
            )
        if not isinstance(end_year, int):
            raise ValueError(f"Input to `end_year`: {end_year}, must be an integer.")
        if end_year < start_year:
            raise ValueError(
                "`start_year`: {start_year}, must less than or equal to the"
                f" `end_year`: {end_year}"
            )

        # Create the date range
        dates = annualized_date_range(
            self.non_operational_start, start_year, self.non_operational_end, end_year
        )
        object.__setattr__(self, "non_operational_dates", dates)
        object.__setattr__(self, "non_operational_dates_set", set(dates))

        # Update the operating dates field to ensure there is no overlap
        if hasattr(self, "operating_dates") and hasattr(self, "non_operational_dates"):
            operating = np.setdiff1d(self.operating_dates, self.non_operational_dates)
            object.__setattr__(self, "operating_dates", operating)
            object.__setattr__(self, "operating_dates_set", set(operating))

    def set_reduced_speed_parameters(
        self,
        start_date: datetime.datetime | None = None,
        start_year: int | None = None,
        end_date: datetime.datetime | None = None,
        end_year: int | None = None,
        speed: float = 0.0,
    ) -> None:
        """Create an annualized list of dates where servicing equipment should be
        operating with reduced speeds. The passed ``start_date``, ``end_date``, and
        ``speed`` will override provided values if they are more conservative than the
        current settings, or a value for ``speed`` will only override the existing value
        if the passed value is slower. Takes a ``start_year`` and ``end_year`` to
        determine how many years should be covered by this setting.

        Parameters
        ----------
        start_date : datetime.datetime | None
            The starting date for the annual date range of reduced speeds, by default
            None.
        start_year : int | None
            The first year that has a reduced speed date range, by default None.
        end_date : datetime.datetime | None
            The ending date for the annual range of reduced speeds, by default None.
        end_year : int | None
            The last year that should have a reduced speed date range, by default None.
        speed : float
            The maximum operating speed under a speed-restricted scenario.

        Raises
        ------
        ValueError
            Raised if the starting and ending dates are the same date.
        """
        # Check that the base dates are valid and override if no values were provided
        # or if the new value is more conservative than the originally provided value
        start_date, end_date = self._compare_dates(
            start_date, end_date, "reduced_speed"
        )
        object.__setattr__(self, "reduced_speed_start", start_date)
        object.__setattr__(self, "reduced_speed_end", end_date)

        if start_date is None or end_date is None:
            object.__setattr__(self, "reduced_speed_dates", np.empty(0, dtype="object"))
            object.__setattr__(self, "reduced_speed_dates_set", set())
            return

        # Check that the input year range is valid
        if not isinstance(start_year, int):
            raise ValueError(
                f"Input to `start_year`: {start_year}, must be an integer."
            )
        if not isinstance(end_year, int):
            raise ValueError(f"Input to `end_year`: {end_year}, must be an integer.")
        if end_year < start_year:
            raise ValueError(
                "`start_year`: {start_year}, must less than or equal to the"
                f" `end_year`: {end_year}"
            )

        # Create the date range
        dates = annualized_date_range(
            self.reduced_speed_start, start_year, self.reduced_speed_end, end_year
        )
        object.__setattr__(self, "reduced_speed_dates", dates)
        object.__setattr__(self, "reduced_speed_dates_set", set(dates))

        # Update the reduced speed if none was originally provided
        assert hasattr(self, "reduced_speed")  # mypy helper
        if speed != 0 and (self.reduced_speed == 0 or speed < self.reduced_speed):
            object.__setattr__(self, "reduced_speed", speed)


@define(frozen=True, auto_attribs=True)
class ScheduledServiceEquipmentData(FromDictMixin, DateLimitsMixin):
    """The data class specification for servicing equipment that will use a
    pre-scheduled basis for returning to site.

    Parameters
    ----------
    name: str
        Name of the piece of servicing equipment.
    equipment_rate: float
        Day rate for the equipment/vessel, in USD.
    n_crews : int
        Number of crew units for the equipment.

        .. note:: the input to this does not matter yet, as multi-crew functionality
            is not yet implemented.

    crew : ServiceCrew
        The crew details, see ``ServiceCrew`` for more information.
    start_month : int
        The day to start operations for the rig and crew.
    start_day : int
        The month to start operations for the rig and crew.
    start_year : int
        The year to start operations for the rig and crew.
    end_month : int
        The month to end operations for the rig and crew.
    end_day : int
        The day to end operations for the rig and crew.
    end_year : int
        The year to end operations for the rig and crew.

        .. note:: if the rig comes annually, then the enter the year for the last year
            that the rig and crew will be available.

    capability : str
        The type of capabilities the equipment contains. Must be one of:

        - RMT: remote (no actual equipment BUT no special implementation)
        - DRN: drone
        - CTV: crew transfer vessel/vehicle
        - SCN: small crane (i.e., field support vessel)
        - LCN: large crane (i.e., heavy lift vessel)
        - CAB: cabling vessel/vehicle
        - DSV: diving support vessel

        Please note that "TOW" is unavailable for scheduled servicing equipment
    mobilization_cost : float
        Cost to mobilize the rig and crew.
    mobilization_days : int
        Number of days it takes to mobilize the equipment.
    speed : float
        Maximum transit speed, km/hr.
    speed_reduction_factor : flaot
        Reduction factor for traveling in inclement weather, default 0. When 0, travel
        is stopped when either `max_windspeed_transport` or `max_waveheight_transport`
        is reached, and when 1, `speed` is used.
    max_windspeed_transport : float
        Maximum windspeed for safe transport, m/s.
    max_windspeed_repair : float
        Maximum windspeed for safe operations, m/s.
    max_waveheight_transport : float
        Maximum waveheight for safe transport, m, default 1000 (land-based).
    max_waveheight_repair : float
        Maximum waveheight for safe operations, m, default 1000 (land-based).
    workday_start : int
        The starting hour of a workshift, in 24 hour time.
    workday_end : int
        The ending hour of a workshift, in 24 hour time.
    crew_transfer_time : float
        The number of hours it takes to transfer the crew from the equipment to the
        system, e.g. how long does it take to transfer the crew from the CTV to the
        turbine, default 0.
    onsite : bool
        Indicator for if the rig and crew are based onsite.

        .. note:: if the rig and crew are onsite be sure that the start and end dates
            represent the first and last day/month of the year, respectively, and the
            start and end years represent the fist and last year in the weather file.

    method : str
        Determines if the ship will do all maximum severity repairs first or do all
        the repairs at one turbine before going to the next, by default severity.
        Should by one of "severity" or "turbine".
    port_distance : int | float
        The distance, in km, the equipment must travel to go between port and site, by
        default 0.
    non_operational_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the environment level, an
        undefined or later starting date will be overridden, by default None.
    non_operational_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the environment level, an
        undefined or earlier ending date will be overridden, by default None.
    reduced_speed_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the environment level, an
        undefined or later starting date will be overridden, by default None.
    reduced_speed_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the environment level, an
        undefined or earlier ending date will be overridden, by default None.
    reduced_speed : float
        The maximum operating speed during the annualized reduced speed operations.
        When defined at the environment level, an undefined or faster value will be
        overridden, by default 0.0.
    """

    name: str = field(converter=str)
    equipment_rate: float = field(converter=float)
    n_crews: int = field(converter=int)
    crew: ServiceCrew = field(converter=ServiceCrew.from_dict)  # type: ignore
    capability: list[str] = field(
        converter=convert_to_list_upper,  # type: ignore
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.in_(VALID_EQUIPMENT),
            iterable_validator=attrs.validators.instance_of(list),
        ),
    )
    speed: float = field(converter=float, validator=greater_than_zero)
    max_windspeed_transport: float = field(converter=float)
    max_windspeed_repair: float = field(converter=float)

    mobilization_cost: float = field(default=0, converter=float)
    mobilization_days: int = field(default=0, converter=int)
    max_waveheight_transport: float = field(default=1000.0, converter=float)
    max_waveheight_repair: float = field(default=1000.0, converter=float)
    workday_start: int = field(default=-1, converter=int, validator=valid_hour)
    workday_end: int = field(default=-1, converter=int, validator=valid_hour)
    crew_transfer_time: float = field(converter=float, default=0.0)
    speed_reduction_factor: float = field(
        default=0.0, converter=float, validator=valid_reduction
    )
    port_distance: float = field(default=0.0, converter=float)
    onsite: bool = field(default=False, converter=bool)
    method: str = field(  # type: ignore
        default="severity",
        converter=[str, str.lower],  # type: ignore
        validator=attrs.validators.in_(["turbine", "severity"]),
    )
    start_month: int = field(
        default=-1, converter=int, validator=attrs.validators.ge(0)
    )
    start_day: int = field(default=-1, converter=int, validator=attrs.validators.ge(0))
    start_year: int = field(default=-1, converter=int, validator=attrs.validators.ge(0))
    end_month: int = field(default=-1, converter=int, validator=attrs.validators.ge(0))
    end_day: int = field(default=-1, converter=int, validator=attrs.validators.ge(0))
    end_year: int = field(default=-1, converter=int, validator=attrs.validators.ge(0))
    strategy: str = field(
        default="scheduled", validator=attrs.validators.in_(["scheduled"])
    )
    operating_dates: np.ndarray = field(factory=list, init=False)
    operating_dates_set: set = field(factory=set, init=False)

    non_operational_start: datetime.datetime = field(default=None, converter=parse_date)
    non_operational_end: datetime.datetime = field(
        default=None, converter=parse_date, validator=check_start_stop_dates
    )
    reduced_speed_start: datetime.datetime = field(default=None, converter=parse_date)
    reduced_speed_end: datetime.datetime = field(
        default=None, converter=parse_date, validator=check_start_stop_dates
    )
    reduced_speed: float = field(default=0, converter=float)
    non_operational_dates: pd.DatetimeIndex = field(factory=list, init=False)
    non_operational_dates_set: pd.DatetimeIndex = field(factory=set, init=False)
    reduced_speed_dates: pd.DatetimeIndex = field(factory=set, init=False)

    def create_date_range(self) -> np.ndarray:
        """Create an ``np.ndarray`` of valid operational dates."""
        start_date = datetime.datetime(
            self.start_year, self.start_month, self.start_day
        )

        # If `onsite` is True, then create the range in one go because there should be
        # no discrepancies in the range.
        # Othewise, create a specified date range between two dates in the year range.
        if self.onsite:
            end_date = datetime.datetime(self.end_year, self.end_month, self.end_day)
            date_range = pd.date_range(start_date, end_date).date
        else:
            date_range = annual_date_range(
                self.start_day,
                self.end_day,
                self.start_month,
                self.end_month,
                self.start_year,
                self.end_year,
            )
        return date_range

    def __attrs_post_init__(self) -> None:
        """Post-initialization."""
        object.__setattr__(self, "operating_dates", self.create_date_range())
        object.__setattr__(self, "operating_dates_set", set(self.operating_dates))


@define(frozen=True, auto_attribs=True)
class UnscheduledServiceEquipmentData(FromDictMixin, DateLimitsMixin):
    """The data class specification for servicing equipment that will use either a
    basis of windfarm downtime or total number of requests serviceable by the equipment.

    Parameters
    ----------
    name: str
        Name of the piece of servicing equipment.
    equipment_rate: float
        Day rate for the equipment/vessel, in USD.
    n_crews : int
        Number of crew units for the equipment.

        .. note: the input to this does not matter yet, as multi-crew functionality
            is not yet implemented.

    crew : ServiceCrew
        The crew details, see ``ServiceCrew`` for more information.
    charter_days : int
        The number of days the servicing equipment can be chartered for.
    capability : str
        The type of capabilities the equipment contains. Must be one of:
        - RMT: remote (no actual equipment BUT no special implementation)
        - DRN: drone
        - CTV: crew transfer vessel/vehicle
        - SCN: small crane (i.e., field support vessel)
        - LCN: large crane (i.e., heavy lift vessel)
        - CAB: cabling vessel/vehicle
        - DSV: diving support vessel
        - TOW: tugboat or towing equipment
        - AHV: anchor handling vessel (tugboat that doesn't trigger tow-to-port)
    speed : float
        Maximum transit speed, km/hr.
    tow_speed : float
        The maximum transit speed when towing, km/hr.

        .. note:: This is only required for when the servicing equipment is tugboat
            enabled for a tow-to-port scenario (capability = "TOW")

    speed_reduction_factor : flaot
        Reduction factor for traveling in inclement weather, default 0. When 0, travel
        is stopped when either `max_windspeed_transport` or `max_waveheight_transport`
        is reached, and when 1, `speed` is used.
    max_windspeed_transport : float
        Maximum windspeed for safe transport, m/s.
    max_windspeed_repair : float
        Maximum windspeed for safe operations, m/s.
    max_waveheight_transport : float
        Maximum waveheight for safe transport, m, default 1000 (land-based).
    max_waveheight_repair : float
        Maximum waveheight for safe operations, m, default 1000 (land-based).
    mobilization_cost : float
        Cost to mobilize the rig and crew, default 0.
    mobilization_days : int
        Number of days it takes to mobilize the equipment, default 0.
    strategy : str
        For any unscheduled maintenance servicing equipment, this determines the
        strategy for dispatching. Should be on of "downtime" or "requests".
    strategy_threshold : str
        For downtime-based scenarios, this is based on the operating level, and should
        be in the range (0, 1). For reqest-based scenarios, this is the maximum number
        of requests that are allowed to build up for any given type of unscheduled
        servicing equipment, should be an integer >= 1.
    workday_start : int
        The starting hour of a workshift, in 24 hour time.
    workday_end : int
        The ending hour of a workshift, in 24 hour time.
    crew_transfer_time : float
        The number of hours it takes to transfer the crew from the equipment to the
        system, e.g. how long does it take to transfer the crew from the CTV to the
        turbine, default 0.
    onsite : bool
        Indicator for if the rig and crew are based onsite.

        .. note:: if the rig and crew are onsite be sure that the start and end dates
            represent the first and last day/month of the year, respectively, and the
            start and end years represent the fist and last year in the weather file.

    method : str
        Determines if the ship will do all maximum severity repairs first or do all
        the repairs at one turbine before going to the next, by default severity.
        Should by one of "severity" or "turbine".
    unmoor_hours : int | float
        The number of hours required to unmoor a floating offshore wind turbine in order
        to tow it to port, by default 0.

        .. note:: Required for the tugboat/towing capability, otherwise unused.

    reconnection_hours : int | float
        The number of hours required to reconnect a floating offshore wind turbine after
        being towed back to site, by default 0.

        .. note:: Required for the tugboat/towing capability, otherwise unused.

    port_distance : int | float
        The distance, in km, the equipment must travel to go between port and site, by
        default 0.
    non_operational_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the environment level or the
        port level, if a tugboat, an undefined or later starting date will be
        overridden, by default None.
    non_operational_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the environment level or the
        port level, if a tugboat, an undefined or earlier ending date will be
        overridden, by default None.
    reduced_speed_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the environment level or the
        port level, if a tugboat, an undefined or later starting date will be
        overridden, by default None.
    reduced_speed_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the environment level or the
        port level, if a tugboat, an undefined or earlier ending date will be
        overridden, by default None.
    reduced_speed : float
        The maximum operating speed during the annualized reduced speed operations.
        When defined at the environment level, an undefined or faster value will be
        overridden, by default 0.0.
    """

    name: str = field(converter=str)
    equipment_rate: float = field(converter=float)
    n_crews: int = field(converter=int)
    crew: ServiceCrew = field(converter=ServiceCrew.from_dict)  # type: ignore
    capability: list[str] = field(
        converter=convert_to_list_upper,  # type: ignore
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.in_(VALID_EQUIPMENT),
            iterable_validator=attrs.validators.instance_of(list),
        ),
    )
    speed: float = field(converter=float, validator=greater_than_zero)
    max_windspeed_transport: float = field(converter=float)
    max_windspeed_repair: float = field(converter=float)
    mobilization_cost: float = field(default=0, converter=float)
    mobilization_days: int = field(default=0, converter=int)
    max_waveheight_transport: float = field(default=1000.0, converter=float)
    max_waveheight_repair: float = field(default=1000.0, converter=float)
    workday_start: int = field(default=-1, converter=int, validator=valid_hour)
    workday_end: int = field(default=-1, converter=int, validator=valid_hour)
    crew_transfer_time: float = field(converter=float, default=0.0)
    speed_reduction_factor: float = field(
        default=0.0, converter=float, validator=valid_reduction
    )
    port_distance: float = field(default=0.0, converter=float)
    onsite: bool = field(default=False, converter=bool)
    method: str = field(  # type: ignore
        default="severity",
        converter=[str, str.lower],  # type: ignore
        validator=attrs.validators.in_(["turbine", "severity"]),
    )
    strategy: str | None = field(
        default="unscheduled",
        converter=clean_string_input,
        validator=attrs.validators.in_(UNSCHEDULED_STRATEGIES),
    )
    strategy_threshold: int | float = field(default=-1, converter=float)
    charter_days: int = field(
        default=-1, converter=int, validator=attrs.validators.gt(0)  # type: ignore
    )
    tow_speed: float = field(default=1, converter=float, validator=greater_than_zero)
    unmoor_hours: int | float = field(default=0, converter=float)
    reconnection_hours: int | float = field(default=0, converter=float)
    non_operational_start: datetime.datetime = field(default=None, converter=parse_date)
    non_operational_end: datetime.datetime = field(default=None, converter=parse_date)
    reduced_speed_start: datetime.datetime = field(default=None, converter=parse_date)
    reduced_speed_end: datetime.datetime = field(default=None, converter=parse_date)
    reduced_speed: float = field(default=0, converter=float)
    non_operational_dates: pd.DatetimeIndex = field(factory=list, init=False)
    non_operational_dates_set: pd.DatetimeIndex = field(factory=set, init=False)
    reduced_speed_dates: pd.DatetimeIndex = field(factory=set, init=False)

    @strategy_threshold.validator  # type: ignore
    def _validate_threshold(
        self,
        attribute: Attribute,  # pylint: disable=W0613
        value: int | float,
    ) -> None:
        """Ensure a valid threshold is provided for a given ``strategy``."""
        if self.strategy == "downtime":
            if value <= 0 or value >= 1:
                raise ValueError(
                    "Downtime-based strategies must have a ``strategy_threshold``",
                    "between 0 and 1, non-inclusive!",
                )
        if self.strategy == "requests":
            if value <= 0:
                raise ValueError(
                    "Requests-based strategies must have a ``strategy_threshold``",
                    "greater than 0!",
                )


@define(frozen=True, auto_attribs=True)
class ServiceEquipmentData(FromDictMixin):
    """Helps to determine the type ServiceEquipment that should be used, based on the
    repair strategy for its operation. See
    :py:class:`~data_classes.ScheduledServiceEquipmentData` or
    :py:class:`~data_classes.UnscheduledServiceEquipmentData` for more details on each
    classifcation.

    Parameters
    ----------
    data_dict : dict
        The dictionary that will be used to create the appropriate ServiceEquipmentData.
        This should contain a field called 'strategy' with either "scheduled" or
        "unscheduled" as a value if strategy is not provided as a keyword argument.
    strategy : str, optional
        Should be one of "scheduled", "requests", "downtime". If nothing is provided,
        the equipment configuration will be checked.

    Raises
    ------
    ValueError
        Raised if ``strategy`` is not one of "scheduled" or "unscheduled".

    Examples
    --------
    The below workflow is how a new data
    :py:class:`~data_classes.ScheduledServiceEquipmentData` object could be created via
    a generic/routinized creation method, and is how the
    :py:class:`~service_equipment.ServiceEquipment`'s ``__init__`` method creates the
    settings data.

    >>> from wombat.core.data_classes import  ServiceEquipmentData
    >>>
    >>> data_dict = {
    >>>     "name": "Crew Transfer Vessel 1",
    >>>     "equipment_rate": 1750,
    >>>     "start_month": 1,
    >>>     "start_day": 1,
    >>>     "end_month": 12,
    >>>     "end_day": 31,
    >>>     "start_year": 2002,
    >>>     "end_year": 2014,
    >>>     "onsite": True,
    >>>     "capability": "CTV",
    >>>     "max_severity": 10,
    >>>     "mobilization_cost": 0,
    >>>     "mobilization_days": 0,
    >>>     "speed": 37.04,
    >>>     "max_windspeed_transport": 99,
    >>>     "max_windspeed_repair": 99,
    >>>     "max_waveheight_transport": 1.5,
    >>>     "max_waveheight_repair": 1.5,
    >>>     "strategy": scheduled,
    >>>     "crew_transfer_time": 0.25,
    >>>     "n_crews": 1,
    >>>     "crew": {
    >>>         "day_rate": 0,
    >>>         "n_day_rate": 0,
    >>>         "hourly_rate": 0,
    >>>         "n_hourly_rate": 0,
    >>>     },
    >>> }
    >>> equipment = ServiceEquipmentData(data_dict).determine_type()
    >>> type(equipment)
    """

    data_dict: dict
    strategy: str | None = field(
        kw_only=True, default=None, converter=clean_string_input
    )

    def __attrs_post_init__(self):
        """Post-initialization."""
        if self.strategy is None:
            object.__setattr__(
                self, "strategy", clean_string_input(self.data_dict["strategy"])
            )
        if self.strategy not in VALID_STRATEGIES:
            raise ValueError(
                f"ServiceEquipment strategy should be one of {VALID_STRATEGIES};"
                f" input: {self.strategy}."
            )

    def determine_type(
        self,
    ) -> ScheduledServiceEquipmentData | UnscheduledServiceEquipmentData:
        """Generate the appropriate ServiceEquipmentData variation.

        Returns
        -------
        Union[ScheduledServiceEquipmentData, UnscheduledServiceEquipmentData]
            The appropriate ``xxServiceEquipmentData`` schema depending on the strategy
            the ``ServiceEquipment`` will use.
        """
        if self.strategy == "scheduled":
            return ScheduledServiceEquipmentData.from_dict(self.data_dict)
        elif self.strategy in UNSCHEDULED_STRATEGIES:
            return UnscheduledServiceEquipmentData.from_dict(self.data_dict)
        else:
            # This should not be able to be reached
            raise ValueError("Invalid strategy provided!")


@define(auto_attribs=True)
class EquipmentMap:
    """Internal mapping for servicing equipment strategy information."""

    strategy_threshold: int | float
    equipment: "ServiceEquipment"  # noqa: disable=F821


@define(auto_attribs=True)
class StrategyMap:
    """Internal mapping for equipment capabilities and their data."""

    CTV: list[EquipmentMap] = field(factory=list)
    SCN: list[EquipmentMap] = field(factory=list)
    LCN: list[EquipmentMap] = field(factory=list)
    CAB: list[EquipmentMap] = field(factory=list)
    RMT: list[EquipmentMap] = field(factory=list)
    DRN: list[EquipmentMap] = field(factory=list)
    DSV: list[EquipmentMap] = field(factory=list)
    TOW: list[EquipmentMap] = field(factory=list)
    AHV: list[EquipmentMap] = field(factory=list)
    is_running: bool = field(default=False, init=False)

    def update(
        self,
        capability: str,
        threshold: int | float,
        equipment: "ServiceEquipment",  # noqa: disable=F821
    ) -> None:
        """Update the strategy mapping between capability types and the
        available ``ServiceEquipment`` objects.

        Parameters
        ----------
        capability : str
            The ``equipment``'s capability.
        threshold : int | float
            The threshold for ``equipment``'s strategy.
        equipment : ServiceEquipment
            The actual ``ServiceEquipment`` object to be logged.

        Raises
        ------
        ValueError
            Raised if there is an invalid capability, though this shouldn't be able to
            be reached.
        """
        # Using a mypy ignore because of an unpatched bug using data classes
        if capability == "CTV":
            self.CTV.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "SCN":
            self.SCN.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "LCN":
            self.LCN.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "CAB":
            self.CAB.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "RMT":
            self.RMT.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "DRN":
            self.DRN.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "DSV":
            self.DSV.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "TOW":
            self.TOW.append(EquipmentMap(threshold, equipment))  # type: ignore
        elif capability == "AHV":
            self.AHV.append(EquipmentMap(threshold, equipment))  # type: ignore
        else:
            # This should not even be able to be reached
            raise ValueError(
                f"Invalid servicing equipment '{capability}' has been provided!"
            )
        self.is_running = True

    def get_mapping(self, capability) -> list[EquipmentMap]:
        """Gets the attribute matching the desired :py:attr:`capability`.

        Parameters
        ----------
        capability : str
            A string matching one of the ``UnscheduledServiceEquipmentData.capability``
            (or scheduled) options.

        Returns
        -------
        list[EquipmentMap]
            Returns the matching mapping of available servicing equipment.
        """
        if capability == "CTV":
            return self.CTV
        if capability == "SCN":
            return self.SCN
        if capability == "LCN":
            return self.LCN
        if capability == "CAB":
            return self.CAB
        if capability == "RMT":
            return self.RMT
        if capability == "DRN":
            return self.DRN
        if capability == "DSV":
            return self.DSV
        if capability == "TOW":
            return self.TOW
        if capability == "AHV":
            return self.AHV
        # This should not even be able to be reached
        raise ValueError(
            f"Invalid servicing equipmen capability '{capability}' has been provided!"
        )

    def move_equipment_to_end(self, capability: str, ix: int) -> None:
        """Moves a used equipment to the end of the mapping list to ensure a broader
        variety of servicing equipment are used throughout a simulation.

        Parameters
        ----------
        capability : str
            A string matching one of ``capability`` options for servicing equipment
            dataclasses.
        ix : int
            The index of the used servicing equipent.
        """
        if capability == "CTV":
            self.CTV.append(self.CTV.pop(ix))
        elif capability == "SCN":
            self.SCN.append(self.SCN.pop(ix))
        elif capability == "LCN":
            self.LCN.append(self.LCN.pop(ix))
        elif capability == "CAB":
            self.CAB.append(self.CAB.pop(ix))
        elif capability == "RMT":
            self.RMT.append(self.RMT.pop(ix))
        elif capability == "DRN":
            self.DRN.append(self.DRN.pop(ix))
        elif capability == "DSV":
            self.DSV.append(self.DSV.pop(ix))
        elif capability == "TOW":
            self.TOW.append(self.TOW.pop(ix))
        elif capability == "AHV":
            self.AHV.append(self.AHV.pop(ix))
        else:
            # This should not even be able to be reached
            raise ValueError(
                f"Invalid servicing equipmen capability {capability} has been provided!"
            )


@define(frozen=True, auto_attribs=True)
class PortConfig(FromDictMixin, DateLimitsMixin):
    """Port configurations for offshore wind power plant scenarios.

    Parameters
    ----------
    name : str
        The name of the port, if multiple are used, then be sure this is unique.
    tugboats : list[str]
        file, or list of files to create the port's tugboats.

        .. note:: Each tugboat is considered to be a tugboat + supporting vessels as
        the primary purpose to tow turbines between a repair port and site.
    n_crews : int
        The number of service crews available to be working on repairs simultaneously;
        each crew is able to service exactly one repair.
    crew : ServiceCrew
        The crew details, see ``ServiceCrew`` for more information.
    max_operations : int
        Total number of turbines the port can handle simultaneously.
    workday_start : int
        The starting hour of a workshift, in 24 hour time.
    workday_end : int
        The ending hour of a workshift, in 24 hour time.
    site_distance : int | float
        Distance, in km, a tugboat has to travel to get between site and port.
    annual_fee : int | float
        The annualized fee for access to the repair port that will be distributed
        monthly in the simulation and accounted for on the first of the month from the
        start of the simulation to the end of the simulation.

        .. note:: Don't include this cost in both this category and either the
        ``FixedCosts.operations_management_administration`` bucket or
        ``FixedCosts.marine_management`` category.

    non_operational_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the port level, an undefined or
        later starting date will be overridden by the environment, and any associated
        tubboats will have this value overridden using the same logic, by default None.
    non_operational_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the port level, an undefined
        or earlier ending date will be overridden by the environment, and any associated
        tubboats will have this value overridden using the same logic, by default None.
    reduced_speed_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the port level, an undefined
        or later starting date will be overridden by the environment, and any associated
        tubboats will have this value overridden using the same logic, by default None.
    reduced_speed_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the port level, an undefined
        or earlier ending date will be overridden by the environment, and any associated
        tubboats will have this value overridden using the same logic, by default None.
    reduced_speed : float
        The maximum operating speed during the annualized reduced speed operations.
        When defined at the port level, an undefined or faster value will be overridden
        by the environment, and any associated tubboats will have this value overridden
        using the same logic, by default 0.0.
    """

    name: str = field(converter=str)
    tugboats: list[str | Path] = field(converter=convert_to_list)
    crew: ServiceCrew = field(converter=ServiceCrew.from_dict)  # type: ignore
    n_crews: int = field(default=1, converter=int)
    max_operations: int = field(default=1, converter=int)
    workday_start: int = field(default=-1, converter=int, validator=valid_hour)
    workday_end: int = field(default=-1, converter=int, validator=valid_hour)
    site_distance: float = field(default=0.0, converter=float)
    annual_fee: float = field(
        default=0, converter=float, validator=greater_than_equal_to_zero
    )
    non_operational_start: datetime.datetime = field(default=None, converter=parse_date)
    non_operational_end: datetime.datetime = field(default=None, converter=parse_date)
    reduced_speed_start: datetime.datetime = field(default=None, converter=parse_date)
    reduced_speed_end: datetime.datetime = field(default=None, converter=parse_date)
    reduced_speed: float = field(default=0, converter=float)
    non_operational_dates: pd.DatetimeIndex = field(factory=set, init=False)
    reduced_speed_dates: pd.DatetimeIndex = field(factory=set, init=False)
    non_operational_dates_set: pd.DatetimeIndex = field(factory=set, init=False)
    reduced_speed_dates_set: pd.DatetimeIndex = field(factory=set, init=False)


@define(frozen=True, auto_attribs=True)
class FixedCosts(FromDictMixin):
    """Fixed costs for operating a windfarm. All values are assumed to be in $/kW/yr.

    Parameters
    ----------
    operations : float
        Non-maintenance costs of operating the project. If a value is provided for this
        attribute, then it will zero out all other values, otherwise it will be set to
        the sum of the remaining values.
    operations_management_administration: float
        Activities necessary to forecast, dispatch, sell, and manage the production of
        power from the plant. Includes both the onsite and offsite personnel, software,
        and equipment to coordinate high voltage equipment, switching, port activities,
        and marine activities.

        .. note:: This should only be used when not breaking down the cost into the
            following categories: ``project_management_administration``,
            ``operation_management_administration``, ``marine_management``, and/or
            ``weather_forecasting``

    project_management_administration : float
        Financial reporting, public relations, procurement, parts and stock management,
        H&SE management, training, subcontracts, and general administration.
    marine_management : float
        Coordination of port equipment, vessels, and personnel to carry out inspections
        and maintenance of generation and transmission equipment.
    weather_forecasting : float
        Daily forecast of metocean conditions used to plan maintenance visits and
        estimate project power production.
    condition_monitoring : float
        Monitoring of SCADA data from wind turbine components to optimize performance
        and identify component faults.
    operating_facilities : float
        Co-located offices, parts store, quayside facilities, helipad, refueling
        facilities, hanger (if necesssary), etc.
    environmental_health_safety_monitoring : float
        Coordination and monitoring to ensure compliance with HSE requirements during
        operations.
    insurance : float
        Insurance policies during operational period including All Risk Property,
        Buisness Interuption, Third Party Liability, and Brokers Fee, and Storm
        Coverage.

        .. note:: This should only be used when not breaking down the cost into the
            following categories: ``brokers_fee``, ``operations_all_risk``,
            ``business_interruption``, ``third_party_liability``, and/or
            ``storm_coverage``

    brokers_fee : float
        Fees for arranging the insurance package.
    operations_all_risk : float
        All Risk Property (physical damage). Sudden and unforseen physical loss or
        physical damage to teh plant/assets during the operational phase of a project.
    business_interruption : float
        Sudden and unforseen loss or physical damage to the plant/assets during the
        operational phase of a project causing an interruption.
    third_party_liability : float
        Liability imposed by law, and/or Express Contractual Liability, for bodily
        injury or property damage.
    storm_coverage : float
        Coverage from huricane and tropical storm events (tyipcally for Atlantic Coast
        projects).
    annual_leases_fees : float
        Ongoing payments, including but not limited to: payments to regulatory body for
        permission to operate at project site (terms defined within lease); payments to
        Transmission Systems Operators or Transmission Asseet Owners for rights to
        transport generated power.

        .. note:: This should only be used when not breaking down the cost into the
            following categories: ``submerge_land_lease_costs`` and/or
            ``transmission_charges_rights``

    submerge_land_lease_costs : float
        Payments to submerged land owners for rights to build project during operations.
    transmission_charges_rights : float
        Any payments to Transmissions Systems Operators or Transmission Asset Owners for
        rights to transport generated power.
    onshore_electrical_maintenance : float
        Inspections of cables, transformer, switch gears, power compensation equipment,
        etc. and infrequent repairs

        .. warning:: This should only be used if not modeling these as processes within
            the model. Currently, onshore modeling is not included.

    labor : float
        The costs associated with labor, if not being modeled through the simulated
        processes.
    """

    # Total Cost
    operations: float = field(default=0)

    # Operations, Management, and General Administration
    operations_management_administration: float = field(default=0)
    project_management_administration: float = field(default=0)
    marine_management: float = field(default=0)
    weather_forecasting: float = field(default=0)
    condition_monitoring: float = field(default=0)

    operating_facilities: float = field(default=0)
    environmental_health_safety_monitoring: float = field(default=0)

    # Insurance
    insurance: float = field(default=0)
    brokers_fee: float = field(default=0)
    operations_all_risk: float = field(default=0)
    business_interruption: float = field(default=0)
    third_party_liability: float = field(default=0)
    storm_coverage: float = field(default=0)

    # Annual Leases and Fees
    annual_leases_fees: float = field(default=0)
    submerge_land_lease_costs: float = field(default=0)
    transmission_charges_rights: float = field(default=0)

    onshore_electrical_maintenance: float = field(default=0)

    labor: float = field(default=0)

    resolution: dict = field(init=False)
    hierarchy: dict = field(init=False)

    def cost_category_validator(self, name: str, sub_names: list[str]):
        """Either updates the higher-level cost to be the sum of the category's
        lower-level costs or uses a supplied higher-level cost and zeros out the
        lower-level costs.

        Parameters
        ----------
        name : str
            The higher-level cost category's name.
        sub_names : List[str]
            The lower-level cost names associated with the higher-level category.
        """
        if getattr(self, name) <= 0:
            object.__setattr__(self, name, fsum(getattr(self, el) for el in sub_names))
        else:
            for cost in sub_names:
                object.__setattr__(self, cost, 0)

    def __attrs_post_init__(self):
        """Post-initialization."""
        grouped_names = {
            "operations_management_administration": [
                "project_management_administration",
                "marine_management",
                "weather_forecasting",
                "condition_monitoring",
            ],
            "insurance": [
                "brokers_fee",
                "operations_all_risk",
                "business_interruption",
                "third_party_liability",
                "storm_coverage",
            ],
            "annual_leases_fees": [
                "submerge_land_lease_costs",
                "transmission_charges_rights",
            ],
        }
        individual_names = [
            "operating_facilities",
            "environmental_health_safety_monitoring",
            "onshore_electrical_maintenance",
            "labor",
        ]

        # Check each category for aggregation
        for _name, _sub_names in grouped_names.items():
            self.cost_category_validator(_name, _sub_names)

        # Check for single value aggregation
        self.cost_category_validator("operations", [*grouped_names] + individual_names)

        # Create the value resolution mapping
        object.__setattr__(
            self,
            "resolution",
            {
                "low": ["operations"],
                "medium": [
                    "operations_management_administration",
                    "insurance",
                    "annual_leases_fees",
                    "operating_facilities",
                    "environmental_health_safety_monitoring",
                    "onshore_electrical_maintenance",
                    "labor",
                ],
                "high": [
                    "project_management_administration",
                    "marine_management",
                    "weather_forecasting",
                    "condition_monitoring",
                    "brokers_fee",
                    "operations_all_risk",
                    "business_interruption",
                    "third_party_liability",
                    "storm_coverage",
                    "submerge_land_lease_costs",
                    "transmission_charges_rights",
                    "operating_facilities",
                    "environmental_health_safety_monitoring",
                    "onshore_electrical_maintenance",
                    "labor",
                ],
            },
        )

        object.__setattr__(
            self,
            "hierarchy",
            {
                "operations": {
                    "operations_management_administration": [
                        "project_management_administration",
                        "marine_management",
                        "weather_forecasting",
                        "condition_monitoring",
                    ],
                    "insurance": [
                        "brokers_fee",
                        "operations_all_risk",
                        "business_interruption",
                        "third_party_liability",
                        "storm_coverage",
                    ],
                    "annual_leases_fees": [
                        "submerge_land_lease_costs",
                        "transmission_charges_rights",
                    ],
                    "operating_facilities": [],
                    "environmental_health_safety_monitoring": [],
                    "onshore_electrical_maintenance": [],
                    "labor": [],
                }
            },
        )


@define(auto_attribs=True)
class SubString:
    """A list of the upstream connections for a turbine and its downstream connector.

    Parameters
    ----------
    downstream : str
        The downstream turbine/substation connection id.
    upstream : list[str]
        A list of the upstream turbine connections.
    """

    downstream: str
    upstream: list[str]


@define(auto_attribs=True)
class String:
    """All of the connection information for a complete string in a wind farm.

    Parameters
    ----------
    start : str
        The substation's ID (``System.id``)
    upstream_map : dict[str, SubString]
        The dictionary of each turbine ID in the string and it's upstream ``SubString``.
    """

    start: str
    upstream_map: dict[str, SubString]


@define(auto_attribs=True)
class SubstationMap:
    """A mapping of every ``String`` connected to a substation, excluding export
    connections to other substations.

    Parameters
    ----------
    string_starts : list[str]
        A list of every first turbine's ``System.id`` in a string connected to the
        substation.
    string_map : dict[str, String]
        A dictionary mapping each string starting turbine to its ``String`` data.
    downstream : str
        The ``System.id`` of where the export cable leads. This should be the same
        ``System.id`` as the substation for an interconnection point, or another
        connecting substation.
    """

    string_starts: list[str]
    string_map: dict[str, String]
    downstream: str


@define(auto_attribs=True)
class WindFarmMap:
    """A list of the upstream connections for a turbine and its downstream connector.

    Parameters
    ----------
    substation_map : list[str]
        A dictionary mapping of each substation and its ``SubstationMap``.
    export_cables : list[tuple[str, str]]
        A list of the export cable connections.
    """

    substation_map: dict[str, SubstationMap]
    export_cables: list[tuple[str, str]]

    def get_upstream_connections(
        self, substation: str, string_start: str, node: str, return_cables: bool = True
    ) -> list[str] | tuple[list[str], list[str]]:
        """Retrieve the upstream turbines (and optionally cables) within the wind farm
        graph.

        Parameters
        ----------
        substation : str
            The substation's ``System.id``.
        string_start : str
            The ``System.id`` of the first turbine in the string.
        node : str
            The ``System.id`` of the ending node for a cable connection.
        return_cables : bool
            Indicates if the ``Cable.id`` should be generated for each of the turbines,
            by default True.

        Returns
        -------
        list[str] | tuple[list[str], list[str]]
            A list of ``System.id`` for all of the upstream turbines of ``node`` if
            ``cables=False``, otherwise the upstream turbine and the ``Cable.id`` lists
            are returned.
        """
        strings = self.substation_map[substation].string_map
        upstream = strings[string_start].upstream_map
        turbines = upstream[node].upstream
        if return_cables:
            cables = [
                f"cable::{node if i == 0 else turbines[i - 1]}::{t}"
                for i, t in enumerate(turbines)
            ]
            return turbines, cables
        return turbines

    def get_upstream_connections_from_substation(
        self, substation: str, return_cables: bool = True, by_string: bool = True
    ) -> (
        list[str]
        | tuple[list[str], list[str]]
        | list[list[str]]
        | tuple[list[list[str]], list[list[str]]]
    ):
        """Retrieve the upstream turbines (and optionally, cables) connected to a
        py:attr:`substation` in the wind farm graph.

        Parameters
        ----------
        substation : str
            The py:attr:`System.id` for the substation.
        return_cables : bool, optional
            Indicates if the ``Cable.id`` should be generated for each of the turbines,
            by default True
        by_string : bool, optional
            Indicates if the list of turbines (and cables) should be a nested list for
            each string (py:obj:`True`), or as 1-D list (py:obj:`False`), by default
            True.

        Returns
        -------
        list[str] | tuple[list[str], list[str]]
            A list of ``System.id`` for all of the upstream turbines of ``node`` if
            ``return_cables=False``, otherwise the upstream turbine and the ``Cable.id``
            lists are returned. These are bifurcated in lists of lists for each string
            if ``by_string=True``
        """
        turbines, cables = [], []
        substation_map = self.substation_map[substation]
        start_nodes = substation_map.string_starts
        for start_node in start_nodes:
            # Add the starting node of the string and substation-turbine array cable
            _turbines = [start_node]
            _cables = [f"cable::{substation}::{start_node}"]

            # Add the main components of the string
            _t, _c = self.get_upstream_connections(substation, start_node, start_node)
            _turbines.extend(_t)
            _cables.extend(_c)
            turbines.append(_turbines)
            cables.append(_cables)

        if not by_string:
            turbines = [el for string in turbines for el in string]  # type: ignore
            cables = [el for string in cables for el in string]  # type: ignore

        if return_cables:
            return turbines, cables
        return turbines



"""Provides a consistent way to access the provided Dinwoodie and IEA Task 26 data
libraries.

All library data should adhere to the followind directory structure where <library>
signifies the user's input library path:
```
<library>
  ├── project
    ├── config     <- Project-level configuration files
    ├── port       <- Port configuration files
    ├── plant      <- Wind farm layout files
  ├── cables       <- Export and Array cable configuration files
  ├── substations  <- Substation configuration files
  ├── turbines     <- Turbine configuration and power curve files
  ├── vessels      <- Land-based and offshore servicing equipment configuration files
  ├── weather      <- Weather profiles
  ├── results      <- The analysis log files and any saved output data
```
"""

from __future__ import annotations

import os
import re
from typing import Any
from pathlib import Path

import yaml


ROOT = Path(__file__).parents[2].resolve()
DEFAULT_LIBRARY = ROOT / "library"
CODE_COMPARISON = DEFAULT_LIBRARY / "code_comparison"
BASE_CASES = DEFAULT_LIBRARY / "baseline"

DINWOODIE = CODE_COMPARISON / "dinwoodie"
IEA_26 = DEFAULT_LIBRARY / "code_comparison" / "iea26"

OSW_FIXED = BASE_CASES / "offshore_fixed"
LBW = BASE_CASES / "land_based"

library_map = {
    "DINWOODIE": DINWOODIE,
    "IEA_26": IEA_26,
    "IEA26": IEA_26,
    "OSW_FIXED": OSW_FIXED,
    "LBW": LBW,
    "LAND_BASED": LBW,
}

# YAML loader that is able to read scientific notation
custom_loader = yaml.SafeLoader
custom_loader.add_implicit_resolver(
    "tag:yaml.org,2002:float",
    re.compile(
        """^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$""",
        re.X,
    ),
    list("-+0123456789."),
)


def load_yaml(path: str | Path, fname: str | Path) -> Any:
    """Loads and returns the contents of the YAML file.

    Parameters
    ----------
    path : str | Path
        Path to the file to be loaded.
    fname : str | Path
        Name of the file (ending in .yaml) to be loaded.

    Returns
    -------
    Any
        Whatever content is in the YAML file.
    """
    return yaml.load(open(os.path.join(path, fname)), Loader=custom_loader)


def create_library_structure(
    library_path: str | Path, create_init: bool = False
) -> None:
    """Creates the following library structure at ``library_path``. If ``library_path``
    does not exist, then the method will fail.

    ```
    <library_path>
    ├── project
        ├── config     <- Project-level configuration files
        ├── port       <- Port configuration files
        ├── plant      <- Wind farm layout files
    ├── cables       <- Export and Array cable configuration files
    ├── substations  <- Substation configuration files
    ├── turbines     <- Turbine configuration and power curve files
    ├── vessels      <- Land-based and offshore servicing equipment configuration files
    ├── weather      <- Weather profiles
    ├── results      <- The analysis log files and any saved output data
    ```

    Parameters
    ----------
    library_path : str | Path
        The folder under which the subfolder structure should be created.
    create_init : bool
        If True, create "__init__.py" in each of the folders so that python installation
        processes will register the files, and if False, only create the folders, by
        default False.

    Raises
    ------
    FileNotFoundError
        Raised if ``library_path`` is not a directory
    """
    if isinstance(library_path, str):
        library_path = Path(library_path)
    library_path = library_path.resolve(strict=True)
    if create_init:
        (library_path / "__init__.py").touch()

    folders = (
        "project/config",
        "project/port",
        "project/plant",
        "cables",
        "substations",
        "turbines",
        "vessels",
        "weather",
        "results",
    )
    # Make the project/ subfolder structure once and make the rest without parents
    (library_path / folders[0]).mkdir(parents=True, exist_ok=True)
    for folder in folders:
        f = library_path / folder
        f.mkdir(exist_ok=True)
        if create_init:
            f = f / "__init__.py"
            f.touch()



"""General methods for time-based calculations."""

from __future__ import annotations

import datetime

from dateutil.parser import parse


HOURS_IN_DAY = 24
HOURS_IN_YEAR = 8760


def parse_date(value: str | None | datetime.datetime) -> datetime.datetime | None:
    """Thin wrapper for ``dateutil.parser.parse`` that converts string dates and returns
    back None or the original value if it's None or a ``datetime.datetime`` object,
    respectively.

    Parameters
    ----------
    value : str | None | datetime.datetime
        A month/date or month-date formatted string to be converted to a
        ``datetime.datetime`` object, or ``datetime.datetime`` object, or None.

    Returns
    -------
    datetime.datetime | None
        A converted ``datetime.datetime`` object or None.
    """
    if value is None:
        return value
    if isinstance(value, datetime.datetime):
        return value

    # Ensure there is a common comparison year for all datetime values
    return parse(value).replace(year=2022)


def convert_dt_to_hours(diff: datetime.timedelta) -> float:
    """Convert a ``datetime.timedelta`` object to number of hours at the seconds
    resolution.

    Parameters
    ----------
    diff : datetime.timedelta
        The difference between two ``datetime.datetime`` objects.

    Returns
    -------
    float
        Number of hours between to ``datetime.datetime`` objects.
    """
    days = diff.days * 24 if diff.days > 0 else 0
    seconds = diff.seconds / 60 / 60
    return days + seconds


def hours_until_future_hour(dt: datetime.datetime, hour: int) -> float:
    """Number of hours until a future hour in the same day for ``hour`` <= 24,
    otherwise, it is the number of hours until a time in the proceeding days.

    Parameters
    ----------
    dt : datetime.datetime
        Focal datetime.
    hour : int
        Hour that is later in the day, in 24 hour time.

    Returns
    -------
    float
        Number of hours between the two times.
    """
    if hour >= 24:
        days, hour = divmod(hour, 24)

        # Convert to python int
        days = int(days)
        hour = int(hour)

        new_dt = dt + datetime.timedelta(days=days)
        new_dt = new_dt.replace(hour=hour, minute=0, second=0)
    else:
        new_dt = dt.replace(hour=hour, minute=0, second=0)
    diff = new_dt - dt
    return convert_dt_to_hours(diff)


def check_working_hours(
    env_start: int, env_end: int, workday_start: int, workday_end: int
) -> tuple[int, int]:
    """Checks the working hours of a port or servicing equipment, and overrides a
    default (-1) to the environment's settings, otherwise returns back the input hours.

    Parameters
    ----------
    env_start : int
        The starting hour for the environment's shift
    env_end : int
        The ending hour for the environment's shift
    workday_start : int
        The starting hour to be checked.
    workday_end : int
        The ending hour to be checked.

    Returns
    -------
    tuple[int, int]
        The starting and ending hour to be applied back to the port or servicing
        equipment.
    """
    start_is_invalid = workday_start == -1
    end_is_invalid = workday_end == -1

    start = env_start if start_is_invalid else workday_start
    end = env_end if end_is_invalid else workday_end

    return start, end


def calculate_cost(
    duration: int | float, rate: float, n_rate: int = 1, daily_rate: bool = False
) -> float:
    """Calculates the equipment cost, or labor cost for either salaried or hourly
    employees.

    Parameters
    ----------
    duration : int | float
        Length of time, in hours.
    rate : float
        The labor or equipment rate, in $USD/hour.
    n_rate : int
        Total number of of the rate to be applied, more than one if``rate`` is broken
        down by number of individual laborers (if rate is a labor rate), by default 1.
    daily_rate : bool, optional
        Indicator for if the ``rate`` is a daily rate (True), or hourly rate (False), by
        default False.

    Returns
    -------
    float
        The total cost of the labor performed.
    """
    multiplier = duration / HOURS_IN_DAY if daily_rate else duration
    return multiplier * rate * n_rate



"""Tests for wombat/core/library.py."""


from wombat.core.library import load_yaml, library_map


def test_load_yaml():
    """Tests `load_yaml`."""
    # Test loading the offshore substation from the Dinwoodie library
    library = library_map["DINWOODIE"]
    folder = "substations"
    correct_substation = {
        "capacity_kw": 0,
        "capex_kw": 0,
        "transformer": {
            "name": "substation",
            "maintenance": [
                {
                    "description": "n/a",
                    "time": 0,
                    "materials": 0,
                    "service_equipment": "CTV",
                    "frequency": 0,
                }
            ],
            "failures": {
                1: {
                    "scale": 0,
                    "shape": 0,
                    "time": 0,
                    "materials": 0,
                    "service_equipment": ["CTV"],
                    "operation_reduction": 0,
                    "level": 1,
                    "description": "n/a",
                }
            },
        },
    }

    substation = load_yaml(library / folder, "offshore_substation.yaml")
    assert substation == correct_substation

    # Test loading CTV1 from the IEA26 project
    library = library_map["IEA26"]
    folder = "vessels"
    correct_ctv = {
        "name": "Crew Transfer Vessel 1",
        "equipment_rate": 3500,
        "start_month": 1,
        "start_day": 1,
        "end_month": 12,
        "end_day": 31,
        "start_year": 1996,
        "end_year": 2015,
        "onsite": True,
        "capability": "CTV",
        "workday_start": 7,
        "workday_end": 18,
        "max_severity": 10,
        "mobilization_cost": 0,
        "mobilization_days": 0,
        "speed": 40.74,
        "max_windspeed_transport": 16,
        "max_windspeed_repair": 16,
        "max_waveheight_transport": 2,
        "max_waveheight_repair": 2,
        "strategy": "scheduled",
        "crew_transfer_time": 0.25,
        "n_crews": 1,
        "crew": {
            "day_rate": 0,
            "n_day_rate": 0,
            "hourly_rate": 0,
            "n_hourly_rate": 0,
        },
    }

    ctv = load_yaml(library / folder, "ctv1.yaml")
    assert ctv == correct_ctv

    # Test loading CTV1 from the IEA_26 project
    library = library_map["IEA_26"]
    folder = "vessels"
    correct_ctv = {
        "name": "Crew Transfer Vessel 1",
        "equipment_rate": 3500,
        "start_month": 1,
        "start_day": 1,
        "end_month": 12,
        "end_day": 31,
        "start_year": 1996,
        "end_year": 2015,
        "onsite": True,
        "capability": "CTV",
        "workday_start": 7,
        "workday_end": 18,
        "max_severity": 10,
        "mobilization_cost": 0,
        "mobilization_days": 0,
        "speed": 40.74,
        "max_windspeed_transport": 16,
        "max_windspeed_repair": 16,
        "max_waveheight_transport": 2,
        "max_waveheight_repair": 2,
        "strategy": "scheduled",
        "crew_transfer_time": 0.25,
        "n_crews": 1,
        "crew": {
            "day_rate": 0,
            "n_day_rate": 0,
            "hourly_rate": 0,
            "n_hourly_rate": 0,
        },
    }

    ctv = load_yaml(library / folder, "ctv1.yaml")
    assert ctv == correct_ctv



""""Defines the Cable class and cable simulations."""
from __future__ import annotations

from typing import Generator
from itertools import zip_longest

import numpy as np
import simpy

from wombat.core import (
    Failure,
    Maintenance,
    RepairRequest,
    SubassemblyData,
    WombatEnvironment,
)


class Cable:
    """The cable system/asset class.

    Parameters
    ----------
    windfarm : ``wombat.windfarm.Windfarm``
        The ``Windfarm`` object.
    env : WombatEnvironment
        The simulation environment.
    cable_id : str
        The unique identifier for the cable.
    connection_type : str
        The type of cable. Must be one of "array" or "export".
    start_node : str
        The starting point (``system.id``) (turbine or substation) of the cable segment.
    cable_data : dict
        The dictionary defining the cable segment.
    """

    def __init__(
        self,
        windfarm,
        env: WombatEnvironment,
        connection_type: str,
        start_node: str,
        end_node: str,
        cable_data: dict,
        name: str | None = None,
    ) -> None:
        """Initializes the ``Cable`` class.

        Parameters
        ----------
        windfarm : ``wombat.windfarm.Windfarm``
            The ``Windfarm`` object.
        env : WombatEnvironment
            The simulation environment.
        connection_type : str
            One of "export" or "array".
        cable_id : str
            The unique identifier for the cable.
        start_node : str
            The starting point (``system.id``) (turbine or substation) of the cable
            segment.
        end_node : str
            The ending point (``system.id``) (turbine or substation) of the cable
            segment.
        cable_data : dict
            The dictionary defining the cable segment.
        name : str | None
            The name of the cable to use during logging.
        """
        self.env = env
        self.windfarm = windfarm
        self.connection_type = connection_type
        self.start_node = start_node
        self.end_node = end_node
        self.id = f"cable::{start_node}::{end_node}"
        self.system = windfarm.system(start_node)

        if self.connection_type not in ("array", "export"):
            raise ValueError(
                f"Input to `connection_type` for {self.id} must be one of 'array'"
                " or 'export'."
            )

        cable_data = {**cable_data, "system_value": self.system.value}
        self.data = SubassemblyData.from_dict(cable_data)
        self.name = self.data.name if name is None else name

        self.operating_level = 1.0
        self.servicing = self.env.event()
        self.downstream_failure = self.env.event()
        self.broken = self.env.event()

        # Ensure events start as processed and inactive
        self.servicing.succeed()
        self.downstream_failure.succeed()
        self.broken.succeed()

        # TODO: need to get the time scale of a distribution like this
        self.processes = dict(self._create_processes())

    def set_string_details(self, start_node: str, substation: str):
        """Sets the starting turbine for the string to be used for traversing the
        correct upstream connections when resetting after a failure.

        Parameters
        ----------
        start_node : str
            The ``System.id`` for the starting turbine on a string.
        substation : str
            The ``System.id`` for the string's connecting substation.
        """
        self.string_start = start_node
        self.substation = substation

    def finish_setup(self) -> None:
        """Creates the ``upstream_nodes`` and ``upstream_cables`` attributes for use by
        the cable when triggering usptream failures and resetting them after the repair
        is complete.
        """
        wf_map = self.windfarm.wind_farm_map
        if self.connection_type == "array":
            turbines = [self.end_node]
            if self.end_node == self.string_start:
                _turbines, cables = wf_map.get_upstream_connections(
                    self.substation, self.string_start, self.string_start
                )
            else:
                _turbines, cables = wf_map.get_upstream_connections(
                    self.substation, self.string_start, self.end_node
                )
            turbines.extend(_turbines)

        if self.connection_type == "export":
            turbines, cables = wf_map.get_upstream_connections_from_substation(
                self.substation
            )

        self.upstream_nodes = turbines
        self.upstream_cables = cables

    def _create_processes(self):
        """Creates the processes for each of the failure and maintenance types.

        Yields
        ------
        Tuple[Union[str, int], simpy.events.Process]
            Creates a dictionary to keep track of the running processes within the
            subassembly.
        """
        for level, failure in self.data.failures.items():
            yield level, self.env.process(self.run_single_failure(failure))

        for i, maintenance in enumerate(self.data.maintenance):
            yield f"m{i}", self.env.process(self.run_single_maintenance(maintenance))

    def interrupt_processes(self) -> None:
        """Interrupts all of the running processes within the subassembly except for the
        process associated with failure that triggers the catastrophic failure.

        Parameters
        ----------
        subassembly : Subassembly
            The subassembly that should have all processes interrupted.
        """
        for _, process in self.processes.items():
            try:
                process.interrupt()
            except RuntimeError:
                # This error occurs for the process halting all other processes.
                pass

    def interrupt_all_subassembly_processes(self) -> None:
        """Thin wrapper for ``interrupt_processes`` for consistent usage with system."""
        self.interrupt_processes()

    def stop_all_upstream_processes(self, failure: Failure | Maintenance) -> None:
        """Stops all upstream turbines and cables from producing power by creating a
        ``env.event()`` for each ``System.cable_failure`` and
        ``Cable.downstream_failure``, respectively. In the case of an export cable, each
        string is traversed to stop the substation and upstream turbines and cables.

        Parameters
        ----------
        failure : Failre
            The ``Failure`` that is causing a string shutdown.
        """
        shared_logging = {
            "agent": self.id,
            "action": "repair request",
            "reason": failure.description,
            "additional": "downstream cable failure",
            "request_id": failure.request_id,
        }
        upstream_nodes = self.upstream_nodes
        upstream_cables = self.upstream_cables
        if self.connection_type == "export":
            # Flatten the list of lists for shutting down the upstream connections
            upstream_nodes = [el for string in upstream_nodes for el in string]
            upstream_cables = [el for string in upstream_cables for el in string]

            # Turn off the subation
            substation = self.windfarm.system(self.end_node)
            substation.cable_failure = self.env.event()
            substation.interrupt_all_subassembly_processes()
            self.env.log_action(
                system_id=self.end_node,
                system_name=substation.name,
                system_ol=substation.operating_level,
                part_ol=np.nan,
                **shared_logging,  # type: ignore
            )

        for t_id, c_id in zip_longest(upstream_nodes, upstream_cables, fillvalue=None):
            assert isinstance(t_id, str)
            turbine = self.windfarm.system(t_id)
            turbine.cable_failure = self.env.event()
            turbine.interrupt_all_subassembly_processes()
            self.env.log_action(
                system_id=t_id,
                system_name=turbine.name,
                system_ol=turbine.operating_level,
                part_ol=np.nan,
                **shared_logging,  # type: ignore
            )

            # If at the end of the string, skip any operations on non-existent cables
            if c_id is None:
                continue
            cable = self.windfarm.cable(c_id)
            cable.downstream_failure = self.env.event()
            cable.interrupt_all_subassembly_processes()
            self.env.log_action(
                system_id=c_id,
                system_name=cable.name,
                part_id=c_id,
                part_name=cable.name,
                system_ol=np.nan,
                part_ol=cable.operating_level,
                **shared_logging,  # type: ignore
            )

    def trigger_request(self, action: Maintenance | Failure):
        """Triggers the actual repair or maintenance logic for a failure or maintenance
        event, respectively.

        Parameters
        ----------
        action : Maintenance | Failure
            The maintenance or failure event that triggers a ``RepairRequest``.
        """
        which = "maintenance" if isinstance(action, Maintenance) else "repair"
        self.operating_level *= 1 - action.operation_reduction

        # Automatically submit a repair request
        # NOTE: mypy is not caught up with attrs yet :(
        repair_request = RepairRequest(  # type: ignore
            system_id=self.id,
            system_name=self.name,
            subassembly_id=self.id,
            subassembly_name=self.name,
            severity_level=action.level,
            details=action,
            cable=True,
            upstream_turbines=self.upstream_nodes,
            upstream_cables=self.upstream_cables,
        )
        repair_request = self.system.repair_manager.register_request(repair_request)
        self.env.log_action(
            system_id=self.id,
            system_name=self.name,
            part_id=self.id,
            part_name=self.name,
            system_ol=self.operating_level,
            part_ol=self.operating_level,
            agent=self.name,
            action=f"{which} request",
            reason=action.description,
            additional=f"severity level {action.level}",
            request_id=repair_request.request_id,
        )

        if action.operation_reduction == 1:
            self.broken = self.env.event()
            self.interrupt_all_subassembly_processes()
            self.stop_all_upstream_processes(action)

        # Remove previously submitted requests as a replacement is required
        if action.replacement:
            _ = self.system.repair_manager.purge_subassembly_requests(
                self.id, self.id, exclude=[repair_request.request_id]
            )
        self.system.repair_manager.submit_request(repair_request)

    def run_single_maintenance(self, maintenance: Maintenance) -> Generator:
        """Runs a process to trigger one type of maintenance request throughout the
        simulation.

        Parameters
        ----------
        maintenance : Maintenance
            A maintenance category.

        Yields
        ------
        simpy.events.Timeout
            Time between maintenance requests.
        """
        while True:
            hours_to_next = maintenance.frequency
            if hours_to_next == 0:
                remainder = self.env.max_run_time - self.env.now
                try:
                    yield self.env.timeout(remainder)
                except simpy.Interrupt:
                    remainder -= self.env.now

            while hours_to_next > 0:
                start = -1  # Ensure an interruption before processing is caught
                try:
                    # If the replacement has not been completed, then wait
                    yield self.servicing & self.downstream_failure & self.broken

                    start = self.env.now
                    yield self.env.timeout(hours_to_next)
                    hours_to_next = 0
                    self.trigger_request(maintenance)
                except simpy.Interrupt:
                    if not self.broken.triggered:
                        # The subassembly had to restart the maintenance cycle
                        hours_to_next = 0
                    else:
                        # A different process failed, so subtract the elapsed time
                        # only if it had started to be processed
                        hours_to_next -= 0 if start == -1 else self.env.now - start

    def run_single_failure(self, failure: Failure) -> Generator:
        """Runs a process to trigger one type of failure repair request throughout the
        simulation.

        Parameters
        ----------
        failure : Failure
            A failure classification.

        Yields
        ------
        simpy.events.Timeout
            Time between failure events that need to request a repair.
        """
        while True:
            hours_to_next = failure.hours_to_next_failure()
            if hours_to_next is None:
                remainder = self.env.max_run_time - self.env.now
                try:
                    yield self.env.timeout(remainder)
                except simpy.Interrupt:
                    remainder -= self.env.now

            assert isinstance(hours_to_next, (int, float))  # mypy helper
            while hours_to_next > 0:  # type: ignore
                start = -1  # Ensure an interruption before processing is caught
                try:
                    yield self.servicing & self.downstream_failure & self.broken

                    start = self.env.now
                    yield self.env.timeout(hours_to_next)
                    hours_to_next = 0
                    self.trigger_request(failure)
                except simpy.Interrupt:
                    if not self.broken.triggered:
                        # Restart after fixing
                        hours_to_next = 0
                    else:
                        # A different process failed, so subtract the elapsed time
                        # only if it had started to be processed
                        hours_to_next -= 0 if start == -1 else self.env.now - start



"""Creates the `Port` class that provies the tow-to-port repair capabilities for
offshore floating wind farms. The `Port` will control a series of tugboats enabled
through the "TOW" capability that get automatically dispatched once a tow-to-port repair
is submitted and a tugboat is available (`ServiceEquipment.at_port`). The `Port` also
controls any mooring repairs through the "AHV" capability, which operates similarly to
the tow-to-port except that it will not be released until the repair is completed, and
operates on a strict shift scheduling basis.
"""

from __future__ import annotations

import logging
from typing import Generator
from pathlib import Path

import numpy as np
import simpy
import pandas as pd
from simpy.events import Process, Timeout
from simpy.resources.store import FilterStore

from wombat.windfarm import Windfarm
from wombat.core.mixins import RepairsMixin
from wombat.core.library import load_yaml
from wombat.utilities.time import hours_until_future_hour
from wombat.core.environment import WombatEnvironment
from wombat.core.data_classes import PortConfig, Maintenance, RepairRequest
from wombat.core.repair_management import RepairManager
from wombat.core.service_equipment import ServiceEquipment


# Numpy random generation initialization
random_generator = np.random.default_rng(seed=42)


class Port(RepairsMixin, FilterStore):
    """The offshore wind base port that operates tugboats and performs tow-to-port
    repairs.

    .. note:: The operating costs for the port are incorporated into the ``FixedCosts``
        functionality in the high-levl cost bucket:
        ``operations_management_administration`` or the more granula cost bucket:
        ``marine_management``

    Parameters
    ----------
    env : WombatEnvironment
        The simulation environment instance.
    windfarm : Windfarm
        The simulation windfarm instance.
    repair_manager : RepairManager
        The simulation repair manager instance.
    config : dict | str | Path
        A path to a YAML object or dictionary encoding the port's configuration
        settings. This will be loaded into a ``PortConfig`` object during
        initialization.

    Attributes
    ----------
    env : WombatEnvironment
        The simulation environment instance.
    windfarm : Windfarm
        The simulation windfarm instance.
    manager : RepairManager
        The simulation repair manager instance.
    settings : PortConfig
        The port's configuration settings, as provided by the user.
    requests_serviced : set[str]
        The set of requests that have already been serviced to ensure there are no
        duplications of labor when splitting out the repair requests to be processed.
    turbine_manager : simpy.Resource
        A SimPy ``Resource`` object that limits the number of turbines that can be towed
        to port, so as not to overload the quayside waters, which is controlled by
        ``settings.max_operations``.
    crew_manager : simpy.Resource
        A SimPy ``Resource`` object that limts the number of repairs that can be
        occurring at any given time, which is controlled by ``settings.n_crews``.
    service_equipment_manager : simpy.FilterStore
        A SimPy ``FilterStore`` object that acts as a coordination system for the
        registered tugboats to tow turbines between port and site. In order to tow
        in either direction they must be filtered by ``ServiceEquipment.at_port``. This
        is generated from the tugboat definitions in ``settings.tugboats``.
    active_repairs : dict[str, dict[str, simpy.events.Event]]
        A nested dictionary of turbines, and its associated request IDs with a SimPy
        ``Event``. The use of events allows them to automatically succeed at the end of
        repairs, and once all repairs are processed on a turbine, the tow-to-site
        process can commence.
    """

    def __init__(
        self,
        env: WombatEnvironment,
        windfarm: Windfarm,
        repair_manager: RepairManager,
        config: dict | str | Path,
    ) -> None:
        super().__init__(env, np.inf)

        self.env = env
        self.windfarm = windfarm
        self.manager = repair_manager
        self.requests_serviced: set[str] = set()
        self.invalid_systems: list[str] = []

        self.manager._register_port(self)

        if not isinstance(config, dict):
            try:
                config = load_yaml(env.data_dir / "project/port", config)
            except FileNotFoundError:
                config = load_yaml(env.data_dir / "repair", config)  # type: ignore
                logging.warning(
                    "DeprecationWarning: In v0.8, all port configurations must be"
                    " located in: '<library>/project/port/"
                )
        assert isinstance(config, dict)
        self.settings = PortConfig.from_dict(config)

        self._check_working_hours(which="env")
        self.settings.set_non_operational_dates(
            self.env.non_operational_start,
            self.env.start_year,
            self.env.non_operational_end,
            self.env.end_year,
        )
        self.settings.set_reduced_speed_parameters(
            self.env.reduced_speed_start,
            self.env.start_year,
            self.env.reduced_speed_end,
            self.env.end_year,
            self.env.reduced_speed,
        )

        # Instantiate the crews, tugboats, and turbine availability
        assert isinstance(self.settings, PortConfig)
        self.turbine_manager = simpy.Resource(env, self.settings.max_operations)
        self.crew_manager = simpy.Resource(env, self.settings.n_crews)

        service_equipment = []
        for t in self.settings.tugboats:
            tugboat = ServiceEquipment(self.env, self.windfarm, repair_manager, t)
            tugboat._register_port(self)
            service_equipment.append(tugboat)

        self.service_equipment_manager = simpy.FilterStore(env, len(service_equipment))
        self.service_equipment_manager.items = service_equipment
        self.active_repairs: dict[str, dict[str, simpy.events.Event]] = {}

        # Create partial functions for the labor and equipment costs for clarity
        self.initialize_cost_calculators(which="port")

        # Run the annualized fee logger
        if self.settings.annual_fee > 0:
            self.env.process(self._log_annual_fee())

    def _log_annual_fee(self):
        """Logs the annual port lease fee on a monthly-basis."""
        assert isinstance(self.settings, PortConfig)
        monthly_fee = self.settings.annual_fee / 12.0
        ix_month_starts = self.env.weather.index.day == 1  # 1st of the month
        ix_month_starts &= self.env.weather.index.hour == 0  # at midnight
        ix_month_starts = np.where(ix_month_starts)[0]
        ix_month_starts = ix_month_starts[ix_month_starts > 0]

        # At time 0 log the first monthly fee
        self.env.log_action(
            agent=self.settings.name,
            action="monthly lease fee",
            reason="port lease",
            equipment_cost=monthly_fee,
        )

        for i, ix_month in enumerate(ix_month_starts):
            # Get the time to the start of the next month
            time_to_next = ix_month if i == 0 else ix_month - ix_month_starts[i - 1]

            # Log the fee at the start of each month at midnight
            yield self.env.timeout(time_to_next)
            self.env.log_action(
                agent=self.settings.name,
                action="monthly lease fee",
                reason="port lease",
                equipment_cost=monthly_fee,
            )

    def repair_single(
        self, request: RepairRequest
    ) -> Generator[Timeout | Process, None, None]:
        """Simulation logic to process a single repair request.

        Parameters
        ----------
        request : RepairRequest
            The submitted repair or maintenance request.
        """
        # Retrieve the actual request
        request = yield self.get(lambda r: r == request)  # type: ignore

        # Request a service crew
        crew_request = self.crew_manager.request()
        yield crew_request  # type: ignore

        # Once a crew is available, process the acutal repair
        # Get the shift parameters
        start_shift = self.settings.workday_start
        end_shift = self.settings.workday_end
        continuous_operations = start_shift == 0 and end_shift == 24

        # Set the default hours to process and remaining hours for the repair
        hours_to_process = hours_remaining = request.details.time

        # Create the shared logging among the processes
        system = self.windfarm.system(request.system_id)
        subassembly = getattr(system, request.subassembly_id)
        shared_logging = {
            "agent": self.settings.name,
            "reason": request.details.description,
            "system_id": system.id,
            "system_name": system.name,
            "part_id": subassembly.id,
            "part_name": subassembly.name,
            "request_id": request.request_id,
        }

        # Continue repairing and waiting a shift until the remaining hours to complete
        # the repair is zero
        while hours_remaining > 0:
            current = self.env.simulation_time

            # Check if the workday is limited by shifts and adjust to stay within shift
            if not continuous_operations:
                hours_to_process = hours_until_future_hour(current, end_shift)

            # Delay until the next shift if we're at the end
            if hours_to_process == 0:
                additional = "end of shift; will resume work in the next shift"
                yield self.env.process(
                    self.wait_until_next_shift(additional=additional, **shared_logging)
                )
                continue

            # Process the repair for the minimum of remaining hours to completion and
            # hours available in the shift
            hours_to_process = min(hours_to_process, hours_remaining)
            yield self.env.process(
                self.process_repair(hours_to_process, request.details, **shared_logging)
            )

            # Decrement the remaining hours and reset the default hours to process back
            # to the remaining repair time
            hours_remaining -= hours_to_process
            hours_to_process = hours_remaining

        # Log the completion of the repair
        action = "maintenance" if isinstance(request.details, Maintenance) else "repair"
        self.env.log_action(
            system_id=system.id,
            part_id=subassembly.id,
            part_name=subassembly.name,
            agent=self.settings.name,
            action=f"{action} complete",
            reason=request.details.description,
            materials_cost=request.details.materials,
            additional="complete",
            request_id=request.request_id,
        )

        # Make the crew available again
        self.crew_manager.release(crew_request)

        self.active_repairs[request.system_id][request.request_id].succeed()
        self.env.process(self.manager.register_repair(request, port=True))

    def transfer_requests_from_manager(
        self, system_id: str
    ) -> None | list[RepairRequest]:
        """Gets all of a given system's repair requests from the simulation's repair
        manager, removes them from that queue, and puts them in the port's queue.

        Parameters
        ----------
        system_id : str
            The ``System.id`` attribute from the system that will be repaired at port.

        Returns
        -------
        None | list[RepairRequest]
            The list of repair requests that need to be completed at port.
        """
        requests = self.manager.get_all_requests_for_system(
            self.settings.name, system_id
        )
        if requests is None:
            return requests

        self.items.extend(requests)
        for request in requests:
            self.env.log_action(
                system_id=request.system_id,
                system_name=request.system_name,
                part_id=request.subassembly_id,
                part_name=request.subassembly_name,
                system_ol=float("nan"),
                part_ol=float("nan"),
                agent=self.settings.name,
                action="requests moved to port",
                reason="at-port repair can now proceed",
                request_id=request.request_id,
            )
            self.active_repairs[system_id][request.request_id] = self.env.event()
        return requests

    def run_repairs(self, system_id: str) -> None:
        """Method that transfers the requests from the repair manager and initiates the
        repair sequence.

        Parameters
        ----------
        system_id : str
            The ``System.id`` that is has been towed to port.
        """
        self.active_repairs[system_id] = {}
        requests = self.transfer_requests_from_manager(system_id)
        if requests is not None:
            self.requests_serviced.update([r.request_id for r in requests])
            [self.env.process(self.repair_single(request)) for request in requests]

    def run_tow_to_port(self, request: RepairRequest) -> Generator[Process, None, None]:
        """The method to initiate a tow-to-port repair sequence.

        The process follows the following following routine:

        1. Request a tugboat from the tugboat resource manager and wait
        2. Runs ``ServiceEquipment.tow_to_port``, which encapsulates the traveling to
            site, unmooring, and return tow with a turbine
        3. Transfers the the turbine's repair log to the port, and gets all available
           crews to work on repairs immediately
        4. Requests a tugboat to return the turbine to site
        5. Runs ``ServiceEquipment.tow_to_site()``, which encapsulates the tow back to
           site, reconnection, resetting the operating status, and returning back to
           port

        Parameters
        ----------
        request : RepairRequest
            The request that initiated the process. This is primarily used for logging
            purposes.

        Yields
        ------
        Generator[Process, None, None]
            The series of events constituting the tow-to-port repairs
        """
        # If the request has already been addressed, return
        if request.request_id in self.requests_serviced:
            return

        # Double check in case a delay causes multiple vessels to be interacting with
        # the same turbine

        # Add the requested system to the list of systems undergoing or registered to be
        # undergoing repairs, so this method can't be run again on the same system
        self.invalid_systems.append(request.system_id)

        # If the system is already undergoing repairs from other servicing equipment,
        # then wait until it's done being serviced
        servicing = self.windfarm.system(request.system_id).servicing

        # Wait for a spot to open up in the port queue
        turbine_request = self.turbine_manager.request()

        yield turbine_request & servicing
        seconds_to_wait, *_ = random_generator.integers(low=0, high=30, size=1) / 3600.0
        yield self.env.timeout(seconds_to_wait)
        yield self.windfarm.system(request.system_id).servicing

        # Halt the turbine before going further to avoid issue with requests being
        # being submitted between now and when the tugboat gets to the turbine
        self.manager.halt_requests_for_system(self.windfarm.system(request.system_id))

        # Request a tugboat to retrieve the turbine
        tugboat = yield self.service_equipment_manager.get(
            lambda x: x.at_port
            and (not x.dispatched)
            and "TOW" in x.settings.capability
        )

        # Check that there is enough time to complete towing, connection, and repairs
        # before starting the process, otherwise, wait until the next operational period
        # TODO: use a more sophisticated guess on timing, other than 20 days
        current = self.env.simulation_time.date()
        check_range = set(
            pd.date_range(current, current + pd.Timedelta(days=20), freq="D").date
        )
        intersection = check_range.intersection(self.settings.non_operational_dates_set)
        if intersection:
            hours_to_next = self.hours_to_next_operational_date(
                start_search_date=max(intersection)
            )
            self.env.log_action(
                agent=self.settings.name,
                action="delay",
                reason="non-operational period",
                additional="waiting for next operational period",
                duration=hours_to_next,
            )
            yield self.env.timeout(hours_to_next)  # type: ignore

        self.requests_serviced.update([request.request_id])
        yield self.env.process(tugboat.run_tow_to_port(request))  # type: ignore

        # Make the tugboat available again
        yield self.service_equipment_manager.put(tugboat)

        # Transfer the repairs to the port queue, which will initiate the repair process
        self.run_repairs(request.system_id)

        # Wait for the repairs to complete
        yield simpy.AllOf(self.env, self.active_repairs[request.system_id].values())

        # Request a tugboat to tow the turbine back to site, and open the turbine queue
        tugboat = yield self.service_equipment_manager.get(
            lambda x: x.at_port
            and (not x.dispatched)
            and "TOW" in x.settings.capability
        )
        self.turbine_manager.release(turbine_request)
        yield self.env.process(tugboat.run_tow_to_site(request))  # type: ignore
        self.invalid_systems.pop(self.invalid_systems.index(request.system_id))

        # Make the tugboat available again
        yield self.service_equipment_manager.put(tugboat)

    def run_unscheduled_in_situ(
        self, request: RepairRequest
    ) -> Generator[Process, None, None]:
        """Runs the in-situ repair processes for port-based servicing equipment such as
        tugboats that will always return back to port, but are not necessarily a feature
        of the windfarm itself, such as a crew transfer vessel.

        Parameters
        ----------
        request : RepairRequest
            The request that triggered the non tow-to-port, but port-based servicing
            equipment repair.

        Yields
        ------
        Generator[Process, None, None]
            The travel and repair processes.
        """
        # If the request has already been addressed, return
        if request.request_id in self.requests_serviced:
            return

        # If the system is already undergoing repairs from other servicing equipment,
        # then wait until it's done being serviced, then double check
        yield self.windfarm.system(request.system_id).servicing
        seconds_to_wait, *_ = random_generator.integers(low=0, high=30, size=1) / 3600.0
        yield self.env.timeout(seconds_to_wait)
        yield self.windfarm.system(request.system_id).servicing

        # Halt the turbine before going further to avoid issue with requests being
        # being submitted between now and when the tugboat gets to the turbine
        self.manager.halt_requests_for_system(self.windfarm.system(request.system_id))
        self.requests_serviced.update([request.request_id])

        # Request a vessel that isn't solely a towing vessel
        vessel = yield self.service_equipment_manager.get(
            lambda x: x.at_port
            and (not x.dispatched)
            and x.settings.capability != ["TOW"]
        )
        assert isinstance(vessel, ServiceEquipment)  # mypy: helper
        request = yield self.manager.get(lambda x: x == request)
        yield self.env.process(vessel.in_situ_repair(request))

        # If the tugboat finished mid-shift, the in-situ repair logic will keep it
        # there, so ensure it returns back to port once it's complete
        if not vessel.at_port:
            yield self.env.process(
                vessel.travel(
                    start="site",
                    end="port",
                    agent=vessel.settings.name,
                    reason=f"{request.details.description} complete",
                    system_id=request.system_id,
                    system_name=request.system_name,
                    part_id=request.subassembly_id,
                    part_name=request.subassembly_name,
                    request_id=request.request_id,
                )
            )

        # Make the tugboat available again
        yield self.service_equipment_manager.put(vessel)



"""The package setup for installations."""

from __future__ import annotations

import codecs
from pathlib import Path

from setuptools import setup, find_packages


HERE = Path(__file__).parent

with open(HERE / "README.md", newline="") as readme:
    long_description = readme.read()
    long_description_content_type = "text/markdown"


def read(relative_path: Path | str) -> str:
    """Reads the `relative_path` file."""
    with codecs.open(str(HERE / relative_path), "r") as fp:
        return fp.read()


def get_version(relative_path: Path | str) -> str:
    """Retreives the version number."""
    for line in read(relative_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


description = "Windfarm operations and maintenance cost-benefit analysis tool"
version = get_version(HERE / "wombat" / "__init__.py")


# Requirements
REQUIRED = [
    "attrs>=21",
    "numpy>=1.21",
    "scipy>=1.8",
    "pandas>=1.4",
    "pyarrow>=10",
    "jupyterlab>=3",
    "simpy>=4.0.1",
    "pyyaml>=6",
    "geopy>=2.3",
    "networkx>=2.7",
    "matplotlib>=3.3",
    "nrel-pysam>=4",
    # required for pre-commit CI
    "types-attrs>=19",
    "types-typed-ast>=1.5",
    "types-PyYAML>=6",
    "types-python-dateutil>=2.8",
]
DEVELOPER = [
    "pre-commit>=2.20",
    "pylint>=2.14",
    "flake8>=5",
    "flake8-docstrings>=1.6",
    "flake8_sphinx_links==0.2.1",
    "black>=22.1",
    "isort>=5.10",
    "pytest>=7",
    "pytest-cov>=4",
    "mypy==0.991",
    "ruff==0.0.246",
]
DOCUMENTATION = [
    "Sphinx==4.*",
    "myst-nb>=0.16",
    "myst-parser>=0.17",
    "sphinx-panels>=0.6",
    "sphinx-book-theme>=0.3.3,<1",
    "sphinxcontrib-spelling>=7",
    "linkify-it-py>=2",
    "sphinxcontrib-bibtex>=2.4",
]
extra_package_requirements = {
    "dev": DEVELOPER,
    "docs": DOCUMENTATION,
    "all": DEVELOPER + DOCUMENTATION,
}


setup(
    name="wombat",
    author="Rob Hammond",
    author_email="rob.hammond@nrel.gov",
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    project_urls={
        "Source": "https://github.com/WISDEM/WOMBAT",
        "Documentation": "https://wisdem.github.io/WOMBAT/",
    },
    classifiers=[
        # Source: https://pypi.org/classifiers
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    package_data={"": ["*.yaml", "*.csv"]},
    install_requires=REQUIRED,
    python_requires=">=3.8, <3.11",
    extras_require=extra_package_requirements,
    test_suite="pytest",
    tests_require=["pytest", "pytest-cov"],
)



"""Tests the Windfarm class."""

from pathlib import Path

import numpy as np
import pytest
import numpy.testing as npt
from networkx.classes.digraph import DiGraph

from wombat.core import RepairManager
from wombat.windfarm import Windfarm
from wombat.windfarm.system import Cable, System

from tests.conftest import (
    EXPORT,
    SUBSTATION,
    VESTAS_V90,
    ARRAY_33KV_240MM,
    ARRAY_33KV_630MM,
)


def test_windfarm_init(env_setup):
    """Tests the setup of ``Windfarm``."""
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout.csv", manager)

    correct_substation_id = ["OSS1"]
    correct_substation = System(env, manager, "OSS1", "OSS1", SUBSTATION, "substation")
    correct_turbine_ids = ["S00T1", "S00T2", "S00T3", "S01T4", "S01T5", "S01T6"]
    correct_turbine_names = ["WTG001", "WTG002", "WTG003", "WTG004", "WTG005", "WTG006"]
    correct_turbine_list = [
        System(env, manager, _id, _name, VESTAS_V90, "turbine")
        for _id, _name in zip(correct_turbine_ids, correct_turbine_names)
    ]
    correct_edge_keys = [
        ("OSS1", "OSS1"),
        ("OSS1", "S00T1"),
        ("S00T1", "S00T2"),
        ("S00T2", "S00T3"),  # string 1
        ("OSS1", "S01T4"),
        ("S01T4", "S01T5"),
        ("S01T5", "S01T6"),  # string 2
    ]
    correct_cable_ids = ["::".join(("cable", *el)) for el in correct_edge_keys]
    correct_cable_list = [
        Cable(windfarm, env, "export", "OSS1", "OSS1", EXPORT),
        Cable(windfarm, env, "array", "OSS1", "S00T1", ARRAY_33KV_630MM),
        Cable(windfarm, env, "array", "S00T1", "S00T2", ARRAY_33KV_240MM),
        Cable(windfarm, env, "array", "S00T2", "S00T3", ARRAY_33KV_240MM),
        Cable(windfarm, env, "array", "OSS1", "S01T4", ARRAY_33KV_630MM),
        Cable(windfarm, env, "array", "S01T4", "S01T5", ARRAY_33KV_240MM),
        Cable(windfarm, env, "array", "S01T5", "S01T6", ARRAY_33KV_240MM),
    ]
    correct_N_edges = 7
    correct_N_nodes = 7
    correct_capacity = VESTAS_V90["capacity_kw"] * len(correct_turbine_ids)
    correct_substation_turbine_map = {
        "OSS1": {
            "turbines": np.array(
                ["S00T1", "S01T4", "S00T2", "S00T3", "S01T5", "S01T6"]
            ),
            "weights": np.full(
                correct_N_nodes - 1, VESTAS_V90["capacity_kw"] / correct_capacity
            ),
        }
    }

    # Basic init items
    assert windfarm.repair_manager == manager
    assert manager.windfarm == windfarm
    assert Path(windfarm.env.operations_log_fname).is_file()

    # Windfarm layout checking: 6 turbines, 1 substation, and 6 cables
    assert windfarm.env == env
    assert isinstance(windfarm.graph, DiGraph)
    assert len(windfarm.graph.nodes) == correct_N_nodes
    assert len(windfarm.graph.edges) == correct_N_edges
    assert sorted(windfarm.graph.nodes) == sorted(
        correct_substation_id + correct_turbine_ids
    )
    assert sorted(windfarm.graph.edges) == sorted(correct_edge_keys)
    assert windfarm.substation_id == correct_substation_id
    assert sorted(windfarm.turbine_id) == sorted(correct_turbine_ids)

    assert windfarm.current_availability == 1
    assert windfarm.capacity == correct_capacity

    assert (
        windfarm.substation_turbine_map.keys() == correct_substation_turbine_map.keys()
    )
    oss1_map = windfarm.substation_turbine_map["OSS1"]
    correct_oss1_map = correct_substation_turbine_map["OSS1"]

    # For the substation map with turbines and their weights, the ordering of the
    # turbines doesn't matter when comparing because the weights are in the same order
    # as the turbines and the capacities in this model are all the same
    npt.assert_equal(sorted(oss1_map["turbines"]), sorted(correct_oss1_map["turbines"]))
    npt.assert_equal(oss1_map["weights"], correct_oss1_map["weights"])

    # TODO: CHECK THE DISTANCE MATRIX
    # TODO: WHERE IS THE LAYOUT DISTANCE FIELD BEING USED IN WINDFARM????
    correct_distances = np.array([0.0, np.inf])  # no distances
    distances = np.unique(windfarm.distance_matrix)
    npt.assert_equal(distances, correct_distances)

    # Check the actual windfarm objects for correctness. This is not a complete check
    # because that is the purpose of test_system.py and test_subassembly.py
    for _id, correct_turbine in zip(correct_turbine_ids, correct_turbine_list):
        turbine = windfarm.graph.nodes[_id]["system"]
        assert turbine.env == correct_turbine.env
        assert turbine.repair_manager == correct_turbine.repair_manager
        assert turbine.id == correct_turbine.id
        assert turbine.name == correct_turbine.name
        assert turbine.capacity == correct_turbine.capacity
        assert turbine.value == correct_turbine.value
        assert turbine.operating_level == correct_turbine.operating_level == 1
        windspeeds = np.arange(0, 30, 0.1)
        npt.assert_equal(turbine.power(windspeeds), correct_turbine.power(windspeeds))

        # Rough check of the subassemblies and maintenance/failure creation
        assert (
            turbine.electrical_system.processes.keys()
            == correct_turbine.electrical_system.processes.keys()
        )
        assert (
            turbine.electronic_control.processes.keys()
            == correct_turbine.electronic_control.processes.keys()
        )
        assert (
            turbine.sensors.processes.keys() == correct_turbine.sensors.processes.keys()
        )
        assert (
            turbine.hydraulic_system.processes.keys()
            == correct_turbine.hydraulic_system.processes.keys()
        )
        assert (
            turbine.yaw_system.processes.keys()
            == correct_turbine.yaw_system.processes.keys()
        )
        assert (
            turbine.rotor_blades.processes.keys()
            == correct_turbine.rotor_blades.processes.keys()
        )
        assert (
            turbine.mechanical_brake.processes.keys()
            == correct_turbine.mechanical_brake.processes.keys()
        )
        assert (
            turbine.rotor_hub.processes.keys()
            == correct_turbine.rotor_hub.processes.keys()
        )
        assert (
            turbine.gearbox.processes.keys() == correct_turbine.gearbox.processes.keys()
        )
        assert (
            turbine.generator.processes.keys()
            == correct_turbine.generator.processes.keys()
        )
        assert (
            turbine.supporting_structure.processes.keys()
            == correct_turbine.supporting_structure.processes.keys()
        )
        assert (
            turbine.drive_train.processes.keys()
            == correct_turbine.drive_train.processes.keys()
        )

    # Check the substation
    substation = windfarm.graph.nodes[correct_substation_id[0]]["system"]
    assert substation.env == correct_substation.env
    assert substation.repair_manager == correct_substation.repair_manager
    assert substation.id == correct_substation.id
    assert substation.name == correct_substation.name
    assert substation.capacity == correct_substation.capacity
    assert substation.value == correct_substation.value
    assert substation.operating_level == correct_substation.operating_level == 1

    # Rough check of the subassemblies and maintenance/failure creation
    assert (
        substation.transformer.processes.keys()
        == correct_substation.transformer.processes.keys()
    )

    # Check the cable initialization
    for key, _id, correct_cable in zip(
        correct_edge_keys, correct_cable_ids, correct_cable_list
    ):
        cable = windfarm.graph.edges[key]["cable"]
        assert cable.env == correct_cable.env
        assert cable.id == correct_cable.id
        assert cable.name == correct_cable.name
        assert cable.system == correct_cable.system
        assert cable.data.name == cable.data.name
        assert cable.operating_level == correct_cable.operating_level == 1
        assert cable.broken.triggered == correct_cable.broken.triggered is True
        assert (
            cable.downstream_failure.triggered
            == correct_cable.downstream_failure.triggered
            is True
        )

        # Rough check of the subassemblies and maintenance/failure creation
        assert cable.processes.keys() == correct_cable.processes.keys()


def test_windfarm_failed_init(env_setup):
    """Tests the failing cases for setup of ``Windfarm`` where no data is provided."""
    env = env_setup
    manager = RepairManager(env)

    # Test for bad substation file
    with pytest.raises(ValueError):
        Windfarm(env, "layout_substation_invalid.csv", manager)

    # Test for bad subassembly file
    with pytest.raises(ValueError):
        Windfarm(env, "layout_subassembly_invalid.csv", manager)

    # Test for bad substation file
    with pytest.raises(ValueError):
        Windfarm(env, "layout_array_invalid.csv", manager)



"""Tests for wombat/utilities/utilities.py, except setup_logger()."""


import datetime

import numpy as np
import pytest
import numpy.testing as nptest

from wombat.utilities.time import convert_dt_to_hours, hours_until_future_hour
from wombat.utilities.logging import format_events_log_message
from wombat.utilities.utilities import IEC_power_curve, _mean


def test_convert_dt_to_hours():
    """Tests `convert_dt_to_hours`."""
    # Test for a clean number of hours
    correct_hours = 10.0
    diff = convert_dt_to_hours(datetime.timedelta(hours=10))
    assert diff == correct_hours

    # Test for a days, hours, and minutes
    correct_hours = 36.5
    diff = convert_dt_to_hours(datetime.timedelta(days=1, hours=12, minutes=30))
    assert diff == correct_hours

    # Test for seconds
    correct_hours = 0.01
    diff = convert_dt_to_hours(datetime.timedelta(seconds=36))
    assert diff == correct_hours


def test_hours_until_future_hour():
    """Tests `hours_until_future_hour`."""
    # Test for a time later in the day, with an exact number of hours
    correct_hours = 10.0
    date = datetime.datetime(2021, 12, 20, 8, 0, 0)
    diff = hours_until_future_hour(date, 18)
    assert diff == correct_hours

    # Test for a time later in the day, with a decimal number of hours
    correct_hours = 10.26
    date = datetime.datetime(2021, 12, 20, 7, 44, 24)
    diff = hours_until_future_hour(date, 18)
    assert diff == correct_hours

    # Test for a time in the next day
    correct_hours = 23.26
    date = datetime.datetime(2021, 12, 20, 7, 44, 24)
    diff = hours_until_future_hour(date, 7)
    assert pytest.approx(diff) == correct_hours  # floating point error at ~16th digit

    # Test for next day, but using hours as 24 + 7 (same result as above)
    correct_hours = 23.26
    date = datetime.datetime(2021, 12, 20, 7, 44, 24)
    diff = hours_until_future_hour(date, 31)
    assert pytest.approx(diff) == correct_hours  # floating point error at ~16th digit

    # Test for multiple days
    correct_hours = 47.26
    date = datetime.datetime(2021, 12, 20, 7, 44, 24)
    diff = hours_until_future_hour(date, 55)
    assert pytest.approx(diff) == correct_hours  # floating point error at ~16th digit


def test_mean():
    """Tests `_mean`."""
    correct_mean = 2.0
    args = [1, 2, 3, 1, 2, 3]

    # Test for individual inputs
    assert _mean(1, 2, 3, 1, 2, 3) == correct_mean

    # Test for unpacking
    assert _mean(*args) == correct_mean


def test_format_events_log_message():
    """Tests `format_events_log_message`."""
    # Test that any floating point value gets updated for to be printed as a float,
    # which is a standard 6 decimal digit value printing.
    message = format_events_log_message(
        simulation_time=datetime.datetime(2021, 12, 20, 9, 45, 0),
        env_time=9.75,
        system_id="unit_tests",
        system_name="unit testing",
        part_id="utilities_tests",
        part_name="utilities testing",
        system_ol=1.0,
        part_ol=1,  # integer will get printed to float
        agent="pytest",
        action="test for errors",
        reason="software quality",
        additional="testing the software for correctness",
        duration=1,  # integer will get printed to float
        request_id="M00004",
        materials_cost=0,  # integer will get printed to float
        hourly_labor_cost=0.0,
        salary_labor_cost=140.00000009,  # 7 decimals wil, so will get clipped 6
        equipment_cost=0,  # integer will get printed to float
    )
    correct_message = (
        "2021-12-20 09:45:00|9.750000|unit_tests|unit testing"
        "|utilities_tests|utilities testing|1.000000|1.000000|pytest"
        "|test for errors|software quality"
        "|testing the software for correctness|1.000000|M00004|na"
        "|0.000000|0.000000|140.000000|0.000000|140.000000|140.000000"
    )
    assert message == correct_message


def test_IEC_power_curve():
    """Tests `IEC_power_curve`."""
    # tests are built from the same as OpenOA
    # power curve source: https://github.com/NREL/turbine-models/blob/master/Offshore/2020ATB_NREL_Reference_15MW_240.csv
    nrel_15mw_wind = np.arange(4, 26)
    nrel_15mw_power = np.array(
        [
            720,
            1239,
            2271,
            3817,
            5876,
            8450,
            11536,
            15000,
            15000,
            15000,
            15000,
            15000,
            15000,
            15000,
            15000,
            15000,
            15000,
            15000,
            15000,
            15000,
            15000,
            1500,
        ]
    )

    cut_in = 4
    cut_out = 25
    curve = IEC_power_curve(
        nrel_15mw_wind,
        nrel_15mw_power,
        windspeed_start=cut_in,
        windspeed_end=cut_out,
        bin_width=1,
    )

    # Create the test data
    test_windspeeds = np.arange(0, 31)
    test_power = curve(test_windspeeds)

    # Test all windspeeds outside of cut-in and cut-out windspeeds produce no power
    should_be_zeros = test_power[
        (test_windspeeds < cut_in) | (test_windspeeds > cut_out)
    ]
    nptest.assert_array_equal(should_be_zeros, np.zeros(should_be_zeros.shape))

    # Test all the valid windspeeds are equal
    valid_power = test_power[(test_windspeeds >= cut_in) & (test_windspeeds <= cut_out)]
    nptest.assert_array_equal(nrel_15mw_power, valid_power)
